#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "MsgPopP.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif

static void ClassPartInitialize(WidgetClass);
static void Initialize(Widget, Widget, ArgList args, Cardinal * num);
static void Destroy(Widget);
static void Resize(Widget);
static void Realize(Widget, XtValueMask*, XSetWindowAttributes *);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);
static void Exposure(Widget widget, XEvent *ev, Region region);

/* resouces */
static XtResource resources[] = {
	{XtNstring, XtCString, XtRString, sizeof(String),
		 XtOffsetOf(MsgPopRec,msgpop.str), XtRImmediate, (XtPointer)NULL},
	{XtNdiffxpos, XtCDiffpos, XtRInt, sizeof(int),
		 XtOffsetOf(MsgPopRec,msgpop.dx), XtRImmediate, (XtPointer)0},
	{XtNdiffypos, XtCDiffpos, XtRInt, sizeof(int),
		 XtOffsetOf(MsgPopRec,msgpop.dy), XtRImmediate, (XtPointer)0},
};

static XtActionsRec actions[] =
{
};

static char translations[] = "";

MsgPopClassRec msgPopClassRec = {
	{
		  /* superclass         */ (WidgetClass) (&overrideShellClassRec),
		 /* class_name         */ "MsgPop",
	     /* size               */ sizeof(MsgPopRec),
		 /* class_initialize   */ NULL,
		 /* class_part_initialize */ ClassPartInitialize, 
  	     /* Class init'ed      */ FALSE,
	     /* initialize         */ Initialize,
		 /* initialize_hook    */ NULL,
		 /* realize            */ Realize,
		 /* actions            */ actions,
		 /* num_actions        */ XtNumber(actions),
	     /* resources          */ resources,
		 /* resource_count     */ XtNumber(resources),
		 /* xrm_class          */ NULLQUARK,
		 /* compress_motion    */ TRUE,
		 /* compress_exposure  */ TRUE,
		 /* compress_enterleave */ FALSE,
		 /* visible_interest   */ FALSE,
		 /* destroy            */ Destroy,
		 /* resize             */ Resize,
		 /* expose             */ Exposure,
		 /* set_values         */ SetValues,
		 /* set_values_hook    */ NULL,
		 /* set_values_almost  */ XtInheritSetValuesAlmost,
		 /* get_values_hook    */ NULL,
		 /* accept_focus       */ NULL,
		 /* intrinsics version */ XtVersion,
		 /* callback offsets   */ NULL,
		 /* tm_table		    */ translations,
		 /* query_geometry	    */ NULL,
		 /* display_accelerator */ NULL,
		 /* extension	      */ NULL
	},{
		/* geometry_manager   */    XtInheritGeometryManager,
 	    /* change_managed     */    XtInheritChangeManaged,
        /* insert_child   */    XtInheritInsertChild,
        /* delete_child   */    XtInheritDeleteChild,
        /* extension      */    NULL
    },{
        /* extension      */    NULL
    },{
        /* extension      */    NULL
    },{
		/* extension 	  */	NULL
	}
};

WidgetClass msgPopWidgetClass = (WidgetClass)&msgPopClassRec;

#define STR			(mpw->msgpop.str)
#define MSGWIN		(mpw->msgpop.msgwin)
#define MINIFRAME   (mpw->msgpop.miniframe)
#define DX			(mpw->msgpop.dx)
#define DY			(mpw->msgpop.dy)

#define OLDSTR			(old->msgpop.str)
#define OLDMSGWIN		(old->msgpop.msgwin)
#define OLDMINIFRAME    (old->msgpop.frame)

static void
ClassPartInitialize(WidgetClass class)
{
	text_fontinfotables_init();
}

static void
Initialize(Widget req, Widget new, ArgList args, Cardinal * num)
{
	MsgPopWidget mpw = (MsgPopWidget)new;
	MSGWIN     = NULL;
	MINIFRAME  = NULL;
	if (STR)
		STR = strdup(STR);
}

static void
Destroy(Widget widget)
{
	MsgPopWidget mpw = (MsgPopWidget)widget;
	if (MINIFRAME)
		miniframe_delete(MINIFRAME);
	if (MSGWIN)
		msgwin_delete(MSGWIN);
	if (STR)
		free(STR);
}

static void
Resize(Widget widget)
{
	MsgPopWidget mpw = (MsgPopWidget)widget;
	if (XtIsRealized(widget) && MSGWIN != NULL &&
		(MSGWIN->w != widget->core.width
		 || MSGWIN->h != widget->core.height)) {
		msgwin_reattach(MSGWIN);
		MSGWIN->shaped = 0;
		msgwin_setpos(MSGWIN, MSGWIN->o_xpos, MSGWIN->o_ypos, DX, DY, 1);
		XtMoveWidget(widget, MSGWIN->xpos, MSGWIN->ypos);
	}
}

static void
StartDrawing(MsgPopWidget mpw)
{
	if (STR) {
		miniframe_open(MINIFRAME, STR);
		XtResizeWidget((Widget)mpw, MAX(MINIFRAME->formatted_width + 
										MSGWIN->radius*4/3 + MSGWIN->lx,
										MSGWIN->radius * 2 + MSGWIN->lx),
					   MAX(MINIFRAME->formatted_height + 
						   MSGWIN->radius*4/3 + MSGWIN->ly,
						   MSGWIN->radius * 2 + MSGWIN->ly), 0);
	}
}

static void
Realize(Widget widget, XtValueMask *value_mask, XSetWindowAttributes *attr)
{
	Display *dpy;
	Window win;
	MsgPopWidget mpw = (MsgPopWidget)widget;
	XtRealizeProc realize;

	realize = msgPopWidgetClass->core_class.superclass->core_class.realize;
	(*realize)(widget, value_mask, attr);

	dpy = XtDisplay(widget);
	win = XtWindow(widget);

	MSGWIN = msgwin_new();
	msgwin_attach_win(MSGWIN, dpy, win);
	msgwin_setpos(MSGWIN, widget->core.x, widget->core.y, DX, DY, 1);
	XtMoveWidget(widget, MSGWIN->xpos, MSGWIN->ypos);
	MINIFRAME  = miniframe_new(dpy, win, 
							   text_get_fitl(dpy), 
							   text_font_name,
							   FONTTYPE_NORMAL, 24,
							   MSGWIN->clip_w);
	StartDrawing(mpw);
}

static Boolean
SetValues(Widget cur, Widget req, Widget new, ArgList args, Cardinal *num_args)
{
	MsgPopWidget old = (MsgPopWidget)cur;
	MsgPopWidget mpw = (MsgPopWidget)new;
	Boolean redisplay = FALSE;

	if (OLDSTR != STR) {
		if (OLDSTR)
			free(OLDSTR);
		if (STR) {
			STR = strdup(STR);
			if (XtIsRealized(cur)) {
				StartDrawing(mpw);
				redisplay = TRUE;
			}
		}
	}
	return redisplay;
}

static void
Exposure(Widget widget, XEvent *ev, Region region)
{
	MsgPopWidget mpw = (MsgPopWidget)widget;
	if (MINIFRAME)
		miniframe_expose_region(MINIFRAME, 
								MSGWIN->off_x, MSGWIN->off_y, region);
}

