#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "xwchar.h"
#include "fonts.h"

/* եȰɽμ */
static char **
GetFontList(Display *dpy, char *name, int *num)
{
	char **fonts;
	int nnames, available;
	for(nnames=1000,available=nnames+1;;) {
		fonts = XListFonts(dpy, name, nnames, &available);
		if (fonts == NULL || available < nnames)
			break;
		XFreeFontNames(fonts);
		nnames = available * 2;
	}
	if (fonts == NULL)
		return NULL;
	*num = available;
	return fonts;
}

static FontInfo *
fontinfo_new(void)
{
	FontInfo *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("fontinfo_new");
		exit(1);
	}
	p->pixels = NULL;
	p->scale_ok = 0;
	p->scalable = 0;
	p->next = NULL;
	return p;
}

static FontData *
fontdata_new(int size)
{
	FontData *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("fontinfo_new");
		exit(1);
	}
	p->size = size;
	p->count = 0;
	p->fs = NULL;
	return p;
}

static void
fontinfo_addpixel(FontInfo *p, int pixel)
{
	FontData *f = p->pixels;
	while (f != NULL) {
		if (f->size == pixel)
			break;
		f = f->next;
	}
	if (f == NULL) {
		f = fontdata_new(pixel);
		f->parent = p;
		f->next = p->pixels;
		p->pixels = f;
	}
}

void
fontinfo_show(FontInfo *p)
{
	char buf[1024];
	sprintf(buf, "-%s-%s-%s-%s---*-*-*-*-%s-*-%s",
			XrmQuarkToString(p->foundry),
			XrmQuarkToString(p->name),
			XrmQuarkToString(p->weight),
			XrmQuarkToString(p->slant),
			XrmQuarkToString(p->spacing),
			XrmQuarkToString(p->encoding));
	fprintf(stderr,"%s\n", buf);
	{
		FontData *f = p->pixels;
		while (f != NULL) {
			fprintf(stderr,"%d ",f->size);
			f = f->next;
		}
	}
	fprintf(stderr,": scaleok[%d] scalable[%d]\n", p->scale_ok, p->scalable);
}

typedef struct FontList
{
	char *name;
} FontList;

static int
compare_font(const void *a, const void *b)
{
    char *p1 = ((FontList*)a)->name, *p2 = ((FontList*)b)->name;
    while (*p1 && *p2 && *p1 == *p2)
        p1++, p2++;
    return(*p1 - *p2);
}

/* Υ󥳡ǥ󥰤եȤξ */
FontInfoList
GetFontInfo(Display *dpy, char *encoding)
{
	char buf[1024];
	char **fonts;
	int fnum;
	int i;

	XrmQuark qencoding = XrmStringToQuark(encoding);
	FontInfo *filist = NULL;
	FontInfo *fi;
	FontList *fl;

	/* get font list and sort */
	sprintf(buf,"-*-*-*-*-*--*-*-*-*-*-*-%s", encoding);
	if ((fonts = GetFontList(dpy, buf, &fnum)) == NULL)
		goto end;
	if ((fl = malloc(sizeof(*fl) * fnum)) == NULL)
		goto end;
	for (i=0;i<fnum;i++)
		fl[i].name = fonts[i];
	qsort(fl, fnum, sizeof(*fl), compare_font);
		
	fi = filist = fontinfo_new();
	for (i=0; i < fnum; i++) {
		int pix, pt, resx, resy, size;
		char foundry[100], name[100], weight[100], slant[100], spacing[100];
		XrmQuark qfoundry, qname, qweight, qslant, qspacing;

		sscanf(fl[i].name, 
			   "-%[^-]-%[^-]-%[^-]-%[^-]-%*[^-]--"
			   "%d-%d-%d-%d-%[^-]-%d-%*s",
			   foundry, name, weight, slant,
			   &pix, &pt, &resx, &resy, spacing, &size);

		qfoundry = XrmStringToQuark(foundry);
		qname    = XrmStringToQuark(name);
		qweight  = XrmStringToQuark(weight);
		qslant   = XrmStringToQuark(slant);
		qspacing = XrmStringToQuark(spacing);

		/* new font */
		if (fi->foundry != qfoundry
			|| fi->name != qname
			|| fi->weight != qweight
			|| fi->slant  != qslant
			|| fi->spacing != qspacing) {

			FontInfo *new = fontinfo_new();
			new->dpy = dpy;
			new->foundry = qfoundry;
			new->name = qname;
			new->weight = qweight;
			new->slant  = qslant;
			new->spacing = qspacing;
			new->encoding = qencoding;
			new->next = fi->next;
			fi->next = new;
			fi = new;
		}

		if (size == 0) {
			if (pt == 0 && resx == 0 && resy == 0)
				fi->scalable = 1;		/* true scalable font */
			else
				fi->scale_ok  = 1;		/* scaling ok */
		} else
			fontinfo_addpixel(fi,pix); 
	}

#if 0
	fi = filist->next;
	while (fi != NULL) {
		fontinfo_show(fi);
		fi = fi->next;
	}
#endif

	free(fl);
 end:
	XFreeFontNames(fonts);
	return filist;
}

static FontInfoTable *
fitable_new(void)
{
	FontInfoTable *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("fitable_new");
		exit(1);
	}
	p->head = 0;
	p->next = NULL;
	return p;
}

/* ƥ󥳡ǥեȤξ */
FontInfoTableList
MakeFontInfoTable(Display *dpy, FontTable *f)
{
	FontInfoTableList fil;
	FontInfoTable *p, *new;
	fil = p = fitable_new();
	while (f->head != 0) {
		FontInfo *fi;
		if ((fi = GetFontInfo(dpy, f->encoding)) != NULL) {
			new = fitable_new();
			new->head = f->head;
			new->info = GetFontInfo(dpy, f->encoding);
			new->next = p->next;
			p->next = new;
			p = new;
		}
		f++;
	}

#if 0
	p = fil->next;
	while (p != NULL) {
		fprintf(stderr,"[%08x]:[%p]\n",p->head,p->info);
		p = p->next;
	}
#endif
	return fil;
}

/* եȤ */
static int
compare_finfo(FontInfo *p, XrmQuark foundry, XrmQuark name,	
			  XrmQuark weight, XrmQuark slant, XrmQuark spacing, int scok)
{
	if (!scok && p->pixels == NULL)
		return -1;
	if (foundry != None && p->foundry != foundry)
		return -1;
	if (name    != None && p->name != name)
		return -1;
	if (weight != None && p->weight != weight)
		return -1;
	if (slant  != None && p->slant  != slant)
		return -1;
	if (spacing != None && p->spacing != spacing)
		return -1;
	return 0;
}

static FontInfo*
compare_finfo2(FontInfoList i, XrmQuark foundry, XrmQuark name,	
			   XrmQuark weight, XrmQuark slant, XrmQuark spacing, int scok)
{
	FontInfo *p = i->next;
	while (p != NULL) {
		if (!compare_finfo(p,foundry,name,weight,slant,spacing,scok))
			return p;
		p = p->next;
	}
	return NULL;
}

/* ŪΥΥեȤõ */
FontInfo *
FindFontInfo(FontInfoList info, 
			 XrmQuark foundry, XrmQuark name,	
			 XrmQuark weight, XrmQuark slant, XrmQuark spacing, int scok)
{
	FontInfo *p;
	/* ɤɤƤ... */
	if ((p = compare_finfo2(info,foundry,name,weight,slant,spacing,scok)))
		return p;
	if ((p = compare_finfo2(info,foundry,name,weight,None,spacing,scok)))
		return p;
	if ((p = compare_finfo2(info,foundry,name,None,None,spacing,scok)))
		return p;
	if ((p = compare_finfo2(info,foundry,name,None,None,None,scok)))
		return p;
	if ((p = compare_finfo2(info,None,name,None,None,None,scok)))
		return p;
	if ((p = compare_finfo2(info,None,None,None,None,None,scok)))
		return p;
	return NULL;	/* ꤨʤϤä */
}

/* ꥵΥեȤμ */
FontData *
FindFontData(FontInfo *i, int pixel, int scale_ok)
{
	FontData *f = i->pixels;
	FontData *r = NULL;
	int a, min = 99999;
	
	/* ֶᤤΤΤõ */
	while (f != NULL) {
		if ((a = abs(f->size - pixel)) == 0) {
			r = f;
			break;
		} else if (a < min) {
			min = a;
			r = f;
		}
		f = f->next;
	}

	if (r)
		r->count++;
	return r;
}
