#ifndef _font_h_
#define _font_h_

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <mywchar.h>

struct FontInfo;

/* font data */
typedef struct FontData {
	struct FontData *next;
	struct FontInfo *parent;
	int size;					/* pixel size      */
	int count;					/* reference count */
	XFontStruct *fs;			/* X font structure */
} FontData;

/* all font data list of the encoding */
typedef struct FontInfo {
	struct FontInfo *next;
	Display *dpy;
	XrmQuark foundry;
	XrmQuark name;		/* font name */
	XrmQuark weight;	/* medium, demibold, bold, black */
	XrmQuark slant;		/* r, i */
	XrmQuark spacing;	/* m, c, p */
	XrmQuark encoding;	/* encoding name */
	FontData *pixels;
	int scale_ok;		/* scaling ok */
	int scalable;		/* true scalable font */
} FontInfo, *FontInfoList;

typedef struct FontTable {
	wchar_t head;
	char *encoding;
	char *encoding2;
	char *encoding3;
} FontTable;

typedef struct FontNameTable {
	wchar_t head;
	XrmQuark foundry;
	XrmQuark name;		/* font name */
} FontNameTable, *FontNameTableList;

typedef struct FontInfoTable {
	struct FontInfoTable *next;
	wchar_t head;
	FontInfoList info;
} FontInfoTable, *FontInfoTableList;

FontInfoList GetFontInfo(Display *dpy, char *encoding);
FontInfoTableList MakeFontInfoTable(Display *dpy, FontTable *tbl);

FontInfo* FindFontInfo(FontInfoList info, XrmQuark foundry, XrmQuark name,	
					   XrmQuark weight, XrmQuark slant, XrmQuark spacing,
					   int scale_ok);
FontData* FindFontData(FontInfo *i, int pixel, int scale_ok);

#endif
