#ifndef _frame_h_
#define _frame_h_

#include "obj.h"
#include "tag.h"
#include "text.h"
#include "timer.h"

typedef struct StyleSheet {
	int data;
} StyleSheet;

#define BR_NORMAL 0x00
#define BR_LEFT   0x01
#define BR_RIGHT  0x02
#define BR_ALL    0x03

typedef enum AlignMode {
	ALIGN_NONE,
	ALIGN_LEFT,
	ALIGN_RIGHT,
	ALIGN_CENTER,
	ALIGN_TOP,
	ALIGN_MIDDLE,
	ALIGN_ABSMIDDLE,
	ALIGN_BOTTOM
} AlignMode;

typedef struct AlignInfo {
	struct AlignInfo *next;
	AlignMode align;
	TAGS tag;
} AlignInfo, *AlignInfoList;

typedef struct FrameInfo {
	struct FrameInfo *next;		/* stack */

	TAGS tag;				/* what tag make this ? */

	/* ե饰 */
	int after_lf;			/* ľե饰           */
	int paragraph;	   		/* ϥե饰           */
	int pre;				/* preformatted flag	    */
	int li_cnt;				/* ꥹѥ		    */
	int li_level;			/* ꥹѥǥȥ٥ */

	AlignInfoList align;	/* ɽΥ饤 */

	int width;	

	int off_x;	/* ɽեå   */
	int off_y;	
	int cur_x;
	int cur_y;

	HTMLObjList left;				
	HTMLObjList right;
	int left_image_margin;	
	int left_image_height;
	int right_image_margin;
	int right_image_height;

	/* Ծ */
	int max_ascent;
	int max_descent;

	int prev_ascent;
	int prev_descent;

} FrameInfo, *FrameInfoList;

typedef struct ColorList {
	struct ColorList *next;
	XrmQuark name;
	u_long pixel;
} ColorList;

typedef struct StyleInfo {
	struct StyleInfo *next;
	TAGS tag;				/* what tag make this ? */
	int size;				/* current font size */
	int fonttype;			/* current font type */
	OC *oc;
	u_long pixel;	/* color pixel */
} StyleInfo, *StyleInfoList;

#define WIDTH  (fi->width - fi->left_image_margin - fi->right_image_margin)
#define CUR_X  (fi->left_image_margin + fi->cur_x)

typedef struct Frame {

	char *name;
	char *path;
	int fd;

	Display *dpy;
	Window win;
	Colormap colormap;
	GC gc;
	Pixmap backup;

	int r_w;	/* ɥμºݤ礭 */
	int r_h;	

	int width;	/* ɽϤ */
	int height;

	int formatted_width;
	int formatted_height;

	int off_x;	/* ɽ offset    */
	int off_y;
	int scr_x;	/* ̻ */
	int scr_y;

	int linkmode;			/* link mode */

	StyleSheet *sheet;

	int basefont;				/* ܥեȥ */
	OC *font[FONTTYPE_ALL][7];	/* եȷ */
	u_long foreground;
	
	StyleInfoList style;		/* ̾ƥȤΥ */
	StyleInfo *link_style;		/* ѤΥ     */

	u_long background;			/* طʿ */
	ImageData *bg_image;		/* ط pixmap */

	FrameInfoList finfo;		/* ǥȾ etc */

	TagWork *tagwork;			/* ѡ   */
	int eof;					/* ѡλե饰 */
	
	HTMLObjList obj;			/* ֥ */
	HTMLObjList line_p;			/* ѡιԤƬΤ */
	HTMLObjList line;			/* ѡιԤƬ */

	HTMLObjList disp_start;
	HTMLObjList disp_end;

	TimerDatList images;		/* ư */

	ColorList *pixels;			/* ο */
	HTMLHrefInfoList links;		/* 󥯾 */
	HTMLNameInfoList names;
} Frame;

Frame *frame_new(Display *dpy, Window win,
				 FontInfoTableList fitl,
				 FontNameTableList fntl,
				 FontNameTableList fntl_m,
				 u_int width, u_int height, int off_x, int off_y);

void frame_resize(Frame *f, int width, int height, int off_x, int off_y);
int  frame_open(Frame *f, int fd, char *path);

void frame_delete(Frame *f);

void frame_expose(Frame *f, int x, int y, int w, int h);
void frame_expose_region(Frame *f, Region reg);

void frame_alloc_color(Frame *f, char *name, u_long *ret);

void frame_push_aligninfo(Frame *f, TAGS tag, AlignMode align);
void frame_pop_aligninfo(Frame *f);
int  frame_find_aligninfo(Frame *f, TAGS tag);

void frame_push_finfo(Frame *f, TAGS tag, int off_x, int off_y);
void frame_pop_finfo(Frame *f);
int  frame_find_finfo(Frame *f, TAGS tag);

StyleInfo *styleinfo_new(Frame *f, TAGS tag);
void styleinfo_setcolor(Frame *f, StyleInfo *p, char *name);
void styleinfo_setsize(Frame *f, StyleInfo *s, int size);
void styleinfo_setfonttype(Frame *f, StyleInfo *s, int type);
void styleinfo_delete(Frame *f, StyleInfo *p);
void frame_push_style(Frame *f, StyleInfo *style);
void frame_pop_style(Frame *f);
int  frame_find_style(Frame *f, TAGS tag);

void frame_add_link(Frame *f, HTMLHrefInfo *p);
void frame_add_obj(Frame *f, HTMLObj *obj);
void frame_add_obj_before_line(Frame *f, HTMLObj *obj);
void frame_add_left(Frame *f, HTMLObj *obj);
void frame_add_right(Frame *f, HTMLObj *obj);
void frame_add_vspace(Frame *f, int linesep);
void frame_add_lf(Frame *f, int mode);

void frame_add_body(Frame *f);
void frame_add_bullet(Frame *f, int type);
void frame_add_string(Frame *f, wchar_t *str, size_t len, int mode);
void frame_add_image(Frame *f, Animate *a, int w, int h, int b, int align);
void frame_add_hr(Frame *f, int align, int width, int size);


/* disp.c */
void frame_check_area(Frame *f);

void frame_disp_all(Frame *f);
void frame_disp_area(Frame *f, int x, int y, int w, int h);
void frame_set_pos(Frame *f, int scr_x, int scr_y);
void frame_scroll(Frame *f, int dx, int dy);
void frame_disp_obj(Frame *f, HTMLObj *p);

FILE *frame_file_open(Frame *f, char *name);

HTMLHrefInfo* frame_get_url(Frame *f, int x, int y);

#endif
