#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "xwchar.h"
#include "parse.h"
#include "text.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif

int
parse_size(char *val, int max)
{
	int ret= 0;
	while (*val){
		if (isdigit(*val)) ret= ret*10 + *val - '0';
		else break;
		val++;
	}
	if (*val=='%' && max) ret= max*ret/100;
	return ret;
}

void
frame_parse_string(Frame *f, Tag *tag)
{
	XCharStruct all;
	wchar_t *start, *wp, *p;
/*	int xlen;*/
	FrameInfo *fi = f->finfo;

	/* in pre tag */
	if (fi->pre == TAG_OPEN) {
		start = p = tag->data;
		while (*p) {
			if (*p == '\n') {
				if (p != start)
					frame_add_string(f, start, p - start, 0);
				else {
					fi->max_ascent  = fi->prev_ascent;
					fi->max_descent = fi->prev_descent;
				}
				frame_add_lf(f,BR_NORMAL);
				start = ++p;
				continue;
			}
			p++;
		}
		if (p != start)
			frame_add_string(f, start, p - start, 0);
		return;
	}

	/* normal string */
	start = wp = p = tag->data;
	while (*start) {
#if 0
		fprintf(stderr,"start[%p] [%p] %d\n",start,fi,fi->cur_x);
#endif
		p = start;
/*		xlen = 0;*/
	word:
		/* paragraph */
		if (fi->paragraph) {
			if (fi->align->align == ALIGN_NONE) {	/*  */
				if (IS_MB(*p))
					fi->cur_x += WcWidth(f->style->oc, *p);
				else	
					fi->cur_x += 20;
			}
			frame_add_vspace(f,10);		/* ϻˤ϶ɲ */
			fi->paragraph = 0;
		}

		/* ñڤФ */
		while (*p && *p != ' ' && !IS_MB(*p)) p++;
/*		xlen += WcWidth(f->style->oc,*p++);*/

		if (*p && (*p == ' ' || wp == p)) p++;
/* 			xlen += WcWidth(f->style->oc,*p++);*/

		WcExtents(f->style->oc, start, p - start, &all);
		if (fi->cur_x + all.rbearing < WIDTH) {

			fi->after_lf = 0;
			/* ǽλ */
			if (!*p) {	
				frame_add_string(f, start, p - start, 0);
				break;
			}
#if 0
			frame_add_string(f, start, p - start, 0);
			wp = start = p;
			continue;
#endif
			wp = p;
			goto word;
		}
			
		/* ľ϶§̵ */
		if (fi->after_lf) {
			frame_add_string(f, start, p - start, 0);
			frame_add_lf(f,BR_NORMAL);
			wp = start = p;
			continue;
		}

		/* § (2ʸޤ) */
		if (wp-1 >= start && gyomatsu_kinsoku(*(wp-1))) {
			wp--;
			if (wp-1 >= start && gyomatsu_kinsoku(*(wp-1)))
				wp--;
		}
		
		if (wp == start) {
			frame_add_lf(f, BR_NORMAL);
			continue;
		}

		/* Ƭ§ (2ʸޤ) */
		if (*wp && gyoto_kinsoku(*wp)) {
			wp++;
			if (*wp && gyoto_kinsoku(*wp))
				wp++;
		}
		
		frame_add_string(f, start, wp - start, 0);
		start = wp;
		frame_add_lf(f, BR_NORMAL);
		continue;
	}
}

static void
frame_parse_heading_open(Frame *f, Tag *tag)
{
	StyleInfo *new = styleinfo_new(f, tag->ti->id);

	if (!f->finfo->after_lf)
		frame_add_lf(f, BR_NORMAL);
	if (f->finfo->tag == TAG_p)
		frame_pop_finfo(f);
	f->finfo->paragraph = 0;

	styleinfo_setsize(f, new, 6 - (tag->ti->id - TAG_h1));
	frame_push_style(f,new);
}

static void
frame_parse_heading_close(Frame *f, Tag *tag)
{
	if (!frame_find_style(f,tag->ti->id)) {
		fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
		return;
	}
	while (f->style->tag != tag->ti->id) {
		if (f->style->next != NULL)
			frame_pop_style(f);
	}
	frame_pop_style(f);
	if (!f->finfo->after_lf)
		frame_add_lf(f, BR_NORMAL);
}

void
frame_parse_heading(Frame *f, Tag *tag)
{
	if (tag->mode == TAG_OPEN)
		frame_parse_heading_open(f,tag);
	else
		frame_parse_heading_close(f,tag);
}

void
frame_parse_block_open(Frame *f, Tag *tag)
{
	/*  */

	if (!f->finfo->after_lf)
		frame_add_lf(f, BR_NORMAL);
	if (f->finfo->align->tag == TAG_p)
		frame_pop_aligninfo(f);
	f->finfo->paragraph = 0;

	switch (tag->ti->id) {
	case TAG_p:
		frame_parse_div(f,tag);
		f->finfo->paragraph = 1;
		break;
	case TAG_ul:
		frame_add_vspace(f,10);
		frame_parse_ul(f,tag);
		break;
	case TAG_ol:
		frame_add_vspace(f,10);
		frame_parse_ol(f,tag);
		break;
	case TAG_dl:
	case TAG_menu:
	case TAG_dir:
		frame_add_vspace(f,10);
		frame_push_finfo(f,tag->ti->id,20,0);
		break;
	case TAG_pre:
		frame_push_finfo(f,tag->ti->id,0,0);
		f->finfo->pre = TAG_OPEN;
		{
			StyleInfo *new = styleinfo_new(f, tag->ti->id);
			styleinfo_setfonttype(f, new, FONTTYPE_PRE);
			frame_push_style(f,new);
		}
		break;
	case TAG_div:
		frame_parse_div(f, tag);
		break;
	case TAG_center:
		frame_push_aligninfo(f, tag->ti->id, ALIGN_CENTER);
		break;
	case TAG_blockquote:
		frame_push_finfo(f,tag->ti->id,20,10);
		break;
	case TAG_hr:
		frame_parse_hr(f,tag);
		break;
	case TAG_form:
	case TAG_isindex:
	case TAG_table:
		break;
	default:
		break;
	}
}

static void
frame_parse_block_close(Frame *f, Tag *tag)
{
	if (f->finfo->align->tag == TAG_p) {
		if (!f->finfo->after_lf)
			frame_add_lf(f, BR_NORMAL);
		frame_pop_aligninfo(f);
	}

	switch (tag->ti->id) {
	case TAG_ul:
	case TAG_ol:
		if (!frame_find_finfo(f,tag->ti->id)) {
			fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
			break;
		}
		while (f->finfo->tag != tag->ti->id) {
			if (f->finfo->next != NULL) {
				if (f->finfo->tag != TAG_li) 
					fprintf(stderr, "forcely close tag [<%d>].\n", 
							f->finfo->tag);
				frame_pop_finfo(f);
			}
		}
		frame_pop_finfo(f);
		break;
	case TAG_dl:
	case TAG_menu:
	case TAG_dir:
		if (!frame_find_finfo(f,tag->ti->id)) {
			fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
			break;
		}
		while (f->finfo->tag != tag->ti->id)
			frame_pop_finfo(f);
		frame_pop_finfo(f);
		break;
	case TAG_p:
		break;
	case TAG_div:
	case TAG_center:
		if (!frame_find_aligninfo(f,tag->ti->id)) {
			fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
			break;
		}
		while (f->finfo->align->tag != tag->ti->id) {
			fprintf(stderr, "forcely close tag [<%d>].\n", f->finfo->tag);
			frame_pop_aligninfo(f);
		}
		frame_pop_aligninfo(f);
		break;
	case TAG_pre:
		if (!frame_find_style(f,tag->ti->id)) {
			fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
		} else {
			while (f->style->tag != tag->ti->id)
				frame_pop_style(f);
			frame_pop_style(f);
		}
		/* no break */
	case TAG_blockquote:
		if (!frame_find_finfo(f,tag->ti->id)) {
			fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
			break;
		}
		while (f->finfo->tag != tag->ti->id) {
			fprintf(stderr, "forcely close tag [<%d>].\n", f->finfo->tag);
			frame_pop_finfo(f);
		}
		frame_pop_finfo(f);
		break;
	case TAG_hr:
		break;
	case TAG_form:
	case TAG_isindex:
	case TAG_table:
		break;
	default:
		break;
	}
}

void
frame_parse_block(Frame *f, Tag *tag)
{
	if (tag->mode == TAG_OPEN)
		frame_parse_block_open(f,tag);
	else
		frame_parse_block_close(f,tag);
}

static char *num_rome[][9] = {
	{"I", "II", "III", "IV", "V", "VI", "VII", "IIX", "IX" },
	{"i", "ii", "iii", "iv", "v", "vi", "vii", "iix", "ix" },
};

void
frame_parse_li(Frame *f, Tag *tag)
{
	switch (tag->ti->id) {
	case TAG_li:
		if (f->finfo->tag == TAG_li)
			frame_pop_finfo(f);
		f->finfo->li_cnt++;
		switch (f->finfo->tag) {
		case TAG_ul:
			frame_add_bullet(f, f->finfo->li_level);
			break;
		case TAG_ol:
			{
				int n;
				char buf[100];
				wchar_t wbuf[100];
				wchar_t *wbuf2;
				mbstate_t s;

				switch (f->finfo->li_level) {
				case 1:
				case 2:
					{
						int n = f->finfo->li_cnt;
						int m, d, c, l, x, i;
						char *p = buf;
						m = n / 1000; n -= m * 1000;
						d = n / 500; n -= d * 500;
						c = n / 100; n -= c * 100;
						l = n / 50;  n -= l * 50;
						x = n / 10;  n -= x * 10;
						for (i=0;i<m;i++)
							*p++ = (f->finfo->li_level == 1)? 'M' : 'm';
						for (i=0;i<d;i++)
							*p++ = (f->finfo->li_level == 1)? 'D' : 'd';
						for (i=0;i<c;i++)
							*p++ = (f->finfo->li_level == 1)? 'C' : 'c';
						for (i=0;i<l;i++)
							*p++ = (f->finfo->li_level == 1)? 'L' : 'l';
						for (i=0;i<x;i++)
							*p++ = (f->finfo->li_level == 1)? 'X' : 'x';
						strcpy(p, num_rome[f->finfo->li_level-1][n-1]);
						strcat(p, ". ");
					}
					break;
				case 3:
					sprintf(buf, "%c. ", 'A' + f->finfo->li_cnt - 1);
					break;
				case 4:
					sprintf(buf, "%c. ", 'a' + f->finfo->li_cnt - 1);
					break;
				default:
					sprintf(buf, "%d. ", f->finfo->li_cnt);
				}
				mbrtowc(NULL,NULL,0,&s);
				n = mbsrtowcs(wbuf, buf, sizeof(buf), &s);
				if ((wbuf2 = malloc(n * sizeof(wbuf2))) != NULL) {
					wcscpy(wbuf2,wbuf);
					frame_add_string(f, wbuf2, n, 1);
					f->obj->prev->type = HTMLOBJ_TEXT_MEM;
				}
			}
			break;
		default:
			break;
		}
		frame_push_finfo(f,tag->ti->id,0,0);
		break;
	default:
	}
}

void
frame_parse_other(Frame *f, Tag *tag)
{
	if (tag->mode == TAG_OPEN) 
		switch (tag->ti->id) {
		case TAG_body:
			frame_parse_body(f,tag);
			break;
		default:
		}	
}

#ifdef DEBUG
void tag_showinfo(Tag *p);
#endif

void
frame_parse(Frame *f)
{
	Tag *tag;
	int ret = 0;
	int id;

	if (f->tagwork == NULL) {
		f->eof = 1;
		return;
	}

	do {
		ret = f->tagwork->main(f->tagwork);
		while ((tag = tagwork_gettag(f->tagwork))) {
#if 0
			tag_showinfo(tag); 
#endif
			switch (tag->type) {
			case TAG_normal:
				id = tag->ti->id;
				if (id >= TAG_b && id <= TAG_textarea)
					frame_parse_text(f,tag);
				else if (id >= TAG_h1 && id <= TAG_h6)
					frame_parse_heading(f,tag);
				else if (id >= TAG_p && id <= TAG_table)
					frame_parse_block(f,tag);
				else if (id >= TAG_lh && id <= TAG_dd)
					frame_parse_li(f,tag);
				else
					frame_parse_other(f,tag);
				break;
			case TAG_string:
				frame_parse_string(f, tag);
				break;
			default:
				break;
			}
		}
	} while (!ret && f->finfo->after_lf == 0);

	if (ret) {
		frame_add_lf(f, BR_ALL);
		while (f->finfo->next)
			frame_pop_finfo(f);
		while (f->style->next)
			frame_pop_style(f);
		f->line = f->obj;
		f->eof = ret;
	}

	f->formatted_height = f->finfo->cur_y;
}
