#include <stdio.h>
#include <stdlib.h>
#include "parse.h"

#include "image.h"
#include "animate.h"

void
frame_parse_head(Frame *f, Tag *tag)
{
}

void
frame_parse_body(Frame *f, Tag *tag)
{
	TagOpt *opt;
	char *bgcol  = NULL;
	char *bgname = NULL;
	char *text   = NULL;

	for (opt = tag->opt; opt; opt=opt->next) {
		switch (opt->oi->id) {
		case OPT_background:
			bgname = opt->val;
			break;
		case OPT_bgcolor:
			bgcol = opt->val;
			break;
		case OPT_text:
			text = opt->val;
			break;
		case OPT_alink:
		case OPT_link:
		case OPT_vlink:
		default:
			break;
		}
	}
#if 0
	fprintf(stderr,"body tag: %s %s\n", bgcol, bgname);
#endif
	if (bgname) {
		ImageData *img;
		FILE *fp;
		if ((fp = frame_file_open(f, bgname)) != NULL) {
			img = image_new();
			if (image_read_stream(img, fp) == 0) {
				image_set_col(img, f->dpy, f->win);
				image_pixmap(img, NULL);
				image_free_data(img);
				f->bg_image = img;
			} else
				image_delete(img);
		}
	}
	if (f->bg_image == NULL && bgcol)
		frame_alloc_color(f, bgcol, &f->background);
	if (text)
		frame_alloc_color(f, text, &(f->style->pixel));
	frame_add_body(f);
}

void
frame_parse_image(Frame *f, Tag *tag)
{
	TagOpt *opt;

	char *name = NULL;
	int align = ALIGN_NONE;

	if (f->finfo->paragraph) {
		if (f->finfo->align->align == ALIGN_NONE) {	/*  */
			f->finfo->cur_x += 20;
		}
		frame_add_vspace(f,10);		/* ϻˤ϶ɲ */
		f->finfo->paragraph = 0;
	}

	for (opt = tag->opt; opt; opt=opt->next) {
		if (!opt->oi) continue;
		switch (opt->oi->id) {
		case OPT_alt:
			break;
		case OPT_width:
			break;
		case OPT_height:
			break;
		case OPT_border:
			break;
		case OPT_align:
			{
				OptInfo *i = GetOptInfo(opt->val);
				if (i) {
					switch (i->id) {
					case OPT_top:
						align = ALIGN_TOP;
						break;
					case OPT_middle:
						align = ALIGN_MIDDLE;
						break;
					case OPT_absmiddle:
						align = ALIGN_ABSMIDDLE;
						break;
					case OPT_bottom:
						align = ALIGN_BOTTOM;
						break;
					case OPT_left:
						align = ALIGN_LEFT;
						break;
					case OPT_right:
						align = ALIGN_RIGHT;
						break;
					default:
						break;
					}
				}
			}
			break;
		case OPT_src:
			name = opt->val;
			break;
		default:
		}
	}
	
	if (!tag->data)
		if (name != NULL) {
			FILE *fp;
			Animate *a;
			a = animate_new();
			if ((fp = frame_file_open(f,name))) {
				if (anim_gif_read_stream(a, fp) == 0) {
					animate_attach_win(a, f->dpy, f->win);
					animate_realize(a);
					animate_go(a);
					tag->data = a;
				} else
					animate_delete(a);
				fclose(fp);
			}
		}

	if (tag->data)
		frame_add_image(f, tag->data, 0, 0, 0, align);
}

void
frame_parse_hr(Frame *f, Tag *tag)
{
	TagOpt *opt;
	FrameInfo *fi = f->finfo;

	int align  = fi->align->align;
	int width  = WIDTH;
	int size   = 2;
	for (opt = tag->opt; opt; opt=opt->next) {
		if (!opt->oi) continue;
		switch (opt->oi->id) {
		case OPT_align:
			{
				OptInfo *i = GetOptInfo(opt->val);
				if (i != NULL) {
					switch (i->id) {
					case OPT_left:
						align = ALIGN_LEFT;
						break;
					case OPT_right:
						align = ALIGN_RIGHT;
						break;
					case OPT_center:
						align = ALIGN_CENTER;
						break;
					default:
						break;
					}
				}
			}
			break;
		case OPT_width:
			width = parse_size(opt->val, width);
			break;
		case OPT_size:
			size = atoi(opt->val);
			break;
		default:
			break;
		}
	}
	frame_add_hr(f, align, width, size);
}

void
frame_parse_anchor(Frame *f, Tag *tag)
{
	TagOpt *opt;

	char *name = NULL;
	char *url  = NULL;

	for (opt = tag->opt; opt; opt=opt->next) {
		if (!opt->oi) continue;
		switch (opt->oi->id) {
		case OPT_href:
			url = opt->val;
			break;
		case OPT_name:
			name = opt->val;
			break;
		default:
			break;
		}
	}

	if (url)
		frame_add_link(f, htmlhrefinfo_new(url));
	
	f->linkmode = 1;
}

void
frame_parse_font(Frame *f, Tag *tag)
{
	TagOpt *opt;
	char *str;
	
	int  size   = 0;
	char *color = NULL;
	StyleInfo *new	= styleinfo_new(f, tag->ti->id);

	for (opt = tag->opt; opt; opt=opt->next) {
		if (!opt->oi) continue;
		switch (opt->oi->id) {
		case OPT_size:
			str = opt->val;
			if( *str == '+' || *str == '-' ){
				size = *(str+1) - '0';
				if (*str == '-') size = -size;
				size += f->basefont;
			} else
				size = *str - '0';
			if (size < 1)
				size = 1;
			else if (size > 7)
				size = 7;
			break;
		case OPT_color:
			color = opt->val;
			break;
		default:
			break;
		}
	}
	
	if (size)
		styleinfo_setsize(f,new,size-1);
	styleinfo_setcolor(f,new,color);
	frame_push_style(f,new);
}


void
frame_parse_div(Frame *f, Tag *tag)
{
	TagOpt *opt;
	int align = f->finfo->align->align;
	for (opt = tag->opt; opt; opt=opt->next) {
		if (!opt->oi) continue;
		switch (opt->oi->id) {
		case OPT_align:
			{
				OptInfo *i = GetOptInfo(opt->val);
				if (i != NULL) {
					switch (i->id) {
					case OPT_left:
						align = ALIGN_LEFT;
						break;
					case OPT_right:
						align = ALIGN_RIGHT;
						break;
					case OPT_center:
						align = ALIGN_CENTER;
						break;
					default:
						break;
					}
				}
			}
			break;
		default:
			break;
		}
	}
	frame_push_aligninfo(f,tag->ti->id,align);
}

void
frame_parse_ul(Frame *f, Tag *tag)
{
	TagOpt *opt;
	int level = f->finfo->li_level + 1;

	for (opt = tag->opt; opt; opt=opt->next) {
		if (!opt->oi) continue;
		switch (opt->oi->id) {
		case OPT_type:
			{
				OptInfo *i = GetOptInfo(opt->val);
				if (i != NULL) {
					switch (i->id) {
					case OPT_disc:
						level = 1;
						break;
					case OPT_square:
						level = 3;
						break;
					case OPT_circle:
						level = 2;
						break;
					default:
						break;
					}
				}
			}
			break;
		default:
			break;
		}
	}
	frame_push_finfo(f,tag->ti->id,20,0);
	f->finfo->li_level = level;
}

void
frame_parse_ol(Frame *f, Tag *tag)
{
	TagOpt *opt;
	int level = 0;
	int cnt   = 0;
	for (opt = tag->opt; opt; opt=opt->next) {
		if (!opt->oi) continue;
		switch (opt->oi->id) {
		case OPT_type:
			if (opt->val) {
				switch (*((char*)opt->val)) {
				case 'I':
					level = 1; break;
				case 'i':
					level = 2; break;
				case 'A':
					level = 3; break;
				case 'a':
					level = 4; break;
				default:
					break;
				}
			}
			break;
		case OPT_start:
			if (opt->val)
				cnt = atoi(opt->val);
			break;
		default:
			break;
		}
	}
	frame_push_finfo(f,tag->ti->id,20,0);
	f->finfo->li_level = level;
	f->finfo->li_cnt   = cnt;
}
