#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "xwchar.h"
#include "parse.h"
#include "text.h"

/* ========================= text markup ================= */

static void
frame_parse_text_open(Frame *f, Tag *tag)
{
	switch (tag->ti->id) {
		/* ========================= font        ================= */
	case TAG_tt:
		break;
	case TAG_i:
	case TAG_b:
		{
			StyleInfo *new = styleinfo_new(f, tag->ti->id);
			styleinfo_setfonttype(f, new, tag->ti->id - TAG_b + FONTTYPE_BOLD);
			frame_push_style(f,new);
		}
		break;
	case TAG_u:
	case TAG_strike:
	case TAG_big:
	case TAG_small:
	case TAG_sub:
	case TAG_sup:
		
		/* ========================= phrase ====================== */
	case TAG_em:
	case TAG_strong:
	case TAG_dfn:
	case TAG_code:
	case TAG_samp:
	case TAG_kbd:
	case TAG_var:
	case TAG_cite:
		break;

		/* ========================= special ===================== */
	case TAG_a:
		frame_parse_anchor(f, tag);
		break;
	case TAG_img:
		frame_parse_image(f,tag);
		break;
	case TAG_font:
		frame_parse_font(f,tag);
		break;
	case TAG_basefont:
		{
			int size = 0;
			TagOpt *opt = tag->opt;
			for(;opt;opt = opt->next) {	
				if (!opt->oi) continue;
				if (opt->oi->id == OPT_size)
					size = atoi(opt->val);
			}
			if (size > 0 && size < 8)
				f->basefont = size;
		}
		break;
	case TAG_br:
		{
			int mode = BR_NORMAL;
			TagOpt *opt = tag->opt;
			for(;opt;opt = opt->next) {	
				if (!opt->oi) continue;
				if (opt->oi->id == OPT_clear) {
					OptInfo *i = GetOptInfo(opt->val);
					if (i) {
						switch (i->id) {
						case OPT_left:
							mode = BR_LEFT;
							break;
						case OPT_right:
							mode = BR_RIGHT;
							break;
						case OPT_all:
							mode = BR_ALL;
							break;
						case OPT_none:
							mode = BR_NORMAL;
							break;
						default:
						}
					}
				}
			}
			frame_add_lf(f, mode);
		}
		break;

	case TAG_applet:
	case TAG_script:
	case TAG_map:
		break;

		/* ==================== form ======================== */

	case TAG_input:
	case TAG_select:
	case TAG_textarea:
		break;
	default:
		break;
	}
}

static void
frame_parse_text_close(Frame *f, Tag *tag)
{
	switch (tag->ti->id) {
	case TAG_tt:
		break;
	case TAG_i:
	case TAG_b:
	case TAG_font:
		if (!frame_find_style(f,tag->ti->id)) {
			fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
			break;
		}
		while (f->style->tag != tag->ti->id)
			frame_pop_style(f);
		frame_pop_style(f);
		break;
	case TAG_u:
	case TAG_strike:
	case TAG_big:
	case TAG_small:
	case TAG_sub:
	case TAG_sup:
		/* ========================= phrase ====================== */
	case TAG_em:
	case TAG_strong:
	case TAG_dfn:
	case TAG_code:
	case TAG_samp:
	case TAG_kbd:
	case TAG_var:
	case TAG_cite:
		break;
	case TAG_a:
		if (f->linkmode)
			f->linkmode = 0;
		else
			fprintf(stderr,"[</%s>] without open tag!\n", tag->ti->name);
		break;
	default:
		break;
	}
}

void
frame_parse_text(Frame *f, Tag *tag)
{
	if (tag->mode == TAG_OPEN)
		frame_parse_text_open(f,tag);
	else
		frame_parse_text_close(f,tag);
}

