#ifndef _tags_h_
#define _tags_h_

#include <sys/types.h>
#include <mywchar.h>

enum {
	TAG_OPEN, TAG_CLOSE
};

typedef enum {
	TAG_none,
	TAG_comment,
	TAG_normal,
	TAG_string,
} TagType;

typedef enum {
	TAG_unknown,

	/* heading */
	TAG_h1, TAG_h2, TAG_h3, TAG_h4, TAG_h5, TAG_h6,

	/* text */
	TAG_b, TAG_i, TAG_tt, TAG_u, TAG_strike,
	TAG_em, TAG_strong, TAG_dfn, TAG_code,
	TAG_samp, TAG_kbd, TAG_var,  TAG_cite, 
	TAG_big, TAG_small, TAG_sub, TAG_sup, TAG_blink, 
	
	TAG_a, TAG_img, TAG_font, TAG_basefont,	TAG_br, 
	TAG_applet, TAG_script, TAG_map,
	TAG_input, TAG_select, TAG_textarea,

	/* block */
	TAG_p,
	TAG_ul, TAG_ol, TAG_dl, TAG_menu, TAG_dir,
	TAG_pre, TAG_div, TAG_center, TAG_blockquote, TAG_hr,
	TAG_form, TAG_isindex, TAG_table, 

	TAG_address,

	TAG_html, TAG_head, TAG_body, TAG_title, TAG_base, TAG_meta,

	/* for list */
	TAG_lh, TAG_li,	TAG_dt, TAG_dd,

	/* for table */
	TAG_caption, TAG_th, TAG_td, TAG_tr,


} TAGS;

typedef enum {
	OPT_unknown,
	OPT_background, OPT_bgcolor, OPT_text, OPT_alink, OPT_link, OPT_vlink,
	OPT_name, OPT_type, OPT_start, OPT_size, OPT_color,
	OPT_noshade,
	OPT_width, OPT_height, OPT_compact,
	OPT_href,
	OPT_src, OPT_lowsrc, OPT_alt, OPT_border,
	OPT_hspace, OPT_vspace,
	OPT_align, OPT_valign,
	OPT_method, OPT_action,
	OPT_value, OPT_maxlength, OPT_checked, OPT_multiple, OPT_selected,
	OPT_rows, OPT_cols, OPT_cellpadding, OPT_cellspacing,
	OPT_nowrap, OPT_rowspan, OPT_colspan,
	OPT_left, OPT_center, OPT_right, 
	OPT_top, OPT_bottom, OPT_middle, OPT_absmiddle,
	OPT_clear, OPT_all, OPT_none,
	OPT_disc, OPT_square, OPT_circle,
} OPTS;

typedef struct TagInfo {
	char *name;
	TAGS id;
	struct TagInfo *next;
} TagInfo;
TagInfo* GetTagInfo(char *s);

typedef struct OptInfo {
	char *name;
	OPTS id;
	struct OptInfo *next;
} OptInfo;
OptInfo *GetOptInfo(char *s);

typedef struct EntList {
    char *name;
    u_short	code;
    struct EntList *next;
} EntList;
EntList *GetEntity(char *s);

typedef struct TagOpt {
	struct TagOpt *next;
	OptInfo *oi;
	char *val;
} TagOpt, *TagOptList;

typedef struct Tag {
	struct Tag *next;
	struct Tag *prev;
	int mode;
	TagType type;
	TagInfo *ti;
	TagOptList opt;
	void *data;
} Tag, *TagList;

typedef struct TagWork {
	int mode;

	int fd;				/* ɤߤǥץ */
	int eof;			/* ɤ߹ߴλե饰		*/

	char *data;			/* multibyte text */
	int pre;			/* pre-formatted mode */
	int tagend;			/* ľե饰 */			

	char *p;			/* current point  */
	char *p2;
	mbstate_t state;	/* state mb->wc   */
	
	TagList tags;		/* tagѡλߥꥹ */
	Tag *cur;			/* html ѡݥ */

	wchar_t (*GetChar)(struct TagWork *);
	int (*main)(struct TagWork *);
} TagWork;

Tag *tagwork_gettag(TagWork *w);
TagWork *tagwork_open(int fd);
void tagwork_delete(TagWork *tw);
void tagwork_reset(TagWork *tw);



#endif
