#include <stdio.h>
#include <sys/types.h>

#include <X11/Xlib.h>

#include "text.h"
#include "xwchar.h"

/* local datas */
static FontTable
fonts[] = {
    { HEAD_ISO8859_1,	"iso8859-1", NULL, NULL },
    { HEAD_JISX0208,	"jisx0208.1983-0", NULL, NULL },
    { HEAD_JISX0201, 	"jisx0201.1976-0", NULL, NULL },
    { HEAD_JISX0212,	"jisx0212.1990-0", NULL, NULL },
	{ HEAD_ISO8859_2,	"iso8859-2", NULL, NULL },
    { HEAD_ISO8859_3,	"iso8859-3", NULL, NULL },
    { HEAD_ISO8859_4,	"iso8859-4", NULL, NULL },
    { HEAD_ISO8859_5,	"iso8859-5", NULL, NULL },
    { HEAD_ISO8859_6,	"iso8859-6", NULL, NULL },
    { HEAD_ISO8859_7,	"iso8859-7", NULL, NULL },
    { HEAD_ISO8859_8,	"iso8859-8", NULL, NULL },
    { HEAD_ISO8859_9,	"iso8859-9", NULL, NULL },
	{ HEAD_KSC5601,		"ksc5601.1987-0", NULL, NULL },
    { 0x62310000,		"viscii1.1-1", NULL, NULL },
    { 0x62540000,   	"tis620.2529-1", NULL, NULL },
    { HEAD_GB2312,		"gb2312.1980-0", NULL, NULL },
    { HEAD_CNS11643_1,	"cns11643.1-0", NULL, NULL },
    { HEAD_CNS11643_2,	"cns11643.2-0", NULL, NULL },
    { HEAD_UNICODE,		"unicode1.1-0", NULL, NULL },
    { 0, NULL,	NULL,	NULL, }
};

/* default font name table */
static struct {
	wchar_t head;
	char *foundry;
	char *name;
} font_data[] = {
    { HEAD_ISO8859_1, "adobe", "times" },
    { HEAD_JISX0208,  "netscape", "fixed" },
    { HEAD_JISX0201,  NULL, "fixed" },
    { HEAD_JISX0212,  NULL, "fixed" },
    { 0,              NULL, NULL }
};

static struct {
	wchar_t head;
	char *foundry;
	char *name;
} font_data2[] = {
    { HEAD_ISO8859_1, NULL, "fixed" },
    { HEAD_JISX0208,  "netscape", "fixed" },
    { HEAD_JISX0201,  NULL, "fixed" },
    { HEAD_JISX0212,  NULL, "fixed" },
    { 0,              NULL, NULL }
};

static
struct FontInfoTables {
	Display *dpy;
	FontInfoTableList fitl;
} fontinfotables[10];

FontNameTable text_font_name[sizeof(font_data)/sizeof(font_data[0])];
FontNameTable text_font_name2[sizeof(font_data2)/sizeof(font_data2[0])];

void
text_fontinfotables_init(void)
{
	static int flag = 1;
	int i;

	if (flag) {
		/* init display - fontinfotable table */
		for (i=0;i<sizeof(fontinfotables)/sizeof(fontinfotables[0]); i++)
			fontinfotables[i].dpy = NULL;

		/* init default font name table */
		for (i=0;i<sizeof(font_data)/sizeof(font_data[0]);i++) {
			text_font_name[i].head = font_data[i].head;
			text_font_name[i].foundry =  (font_data[i].foundry)?
				XrmStringToQuark(font_data[i].foundry):None;
			text_font_name[i].name    =  (font_data[i].name)?
				XrmStringToQuark(font_data[i].name):None;
		}
		for (i=0;i<sizeof(font_data2)/sizeof(font_data2[0]);i++) {
			text_font_name2[i].head    = font_data2[i].head;
			text_font_name2[i].foundry =  (font_data2[i].foundry)?
				XrmStringToQuark(font_data2[i].foundry):None;
			text_font_name2[i].name    =  (font_data2[i].name)?
				XrmStringToQuark(font_data2[i].name):None;
		}
		flag = 0;
	}
}

FontInfoTableList
text_get_fitl(Display *dpy)
{
	struct FontInfoTables *fit = fontinfotables;
	FontInfoTableList fitl;

	/* search same display  == same font */
	while (fit->dpy != NULL) {
		if (fit->dpy == dpy)
			break;
		fit++;
	}

	/* not found !! ... entry new font list to display */
	if (fit->dpy == NULL) {
		fitl = fit->fitl = 	MakeFontInfoTable(dpy, fonts);
		fit->dpy = dpy;
	} else
		fitl = fit->fitl;

	return fitl;
}
