#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>

#include "XMaidP.h"
#include <X11/Xmu/Converters.h>

static void ClassInitialize(void);
static void Initialize(Widget, Widget, ArgList args, Cardinal * num);
static void Destroy(Widget);
static void Realize(Widget, XtValueMask*, XSetWindowAttributes *);
static void ConvertCursor(Widget w);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);

/* private */
static void TimerHandler(XtPointer cl, XtIntervalId *id);
static void TimerHandler2(XtPointer cl, XtIntervalId *id);
static void MoveHandler(XtPointer cl, XtIntervalId *id);
static void Callback(Widget w, XEvent *ev, String *p, Cardinal *n);
static void StartCallback(Widget w, XtPointer , XEvent *ev, Boolean *cont);

static void
XtGetCwd(Widget widget, int offset, XrmValue *value)
{
	char buf[MAXPATHLEN+1];
	value->addr = strdup(getcwd(buf,MAXPATHLEN));
}

#define offset(field) XtOffsetOf(XMaidRec, xmaid.field)
/* resouces */
static XtResource resources[] = {
	{XtNmodulePath, "Path", XtRString, sizeof(String),
		 offset(modulepath), XtRCallProc, (XtPointer)XtGetCwd},
	{XtNmsgPath, "Path", XtRString, sizeof(String),
		 offset(msgpath), XtRCallProc, (XtPointer)XtGetCwd},
	{XtNdataPath, "Path", XtRString, sizeof(String),
		 offset(datapath), XtRCallProc, (XtPointer)XtGetCwd},
	{XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer), 
		 offset(callbacks), XtRCallback, (XtPointer)NULL},
	{XtNstartupCallback, XtCCallback, XtRCallback, sizeof(XtPointer), 
		 offset(startup_callbacks), XtRCallback, (XtPointer)NULL},
	{XtNmoduleName, XtCModuleName, XtRString, sizeof(String),
		 offset(modulename), XtRImmediate, (XtPointer)NULL},
	/* cursors */
	{XtNcursor, XtCCursor, XtRCursor, sizeof(Cursor),
		 offset(cursor), XtRImmediate, (XtPointer) None},
	{XtNcursorName, XtCCursor, XtRString, sizeof(String),
		 offset(cursor_name), XtRString, NULL},
	{XtNpointerColor, XtCForeground, XtRPixel, sizeof(Pixel),
		 offset(pointer_fg), XtRString, XtDefaultForeground},
	{XtNpointerColorBackground, XtCBackground, XtRPixel, sizeof(Pixel),
		 offset(pointer_bg), XtRString, XtDefaultBackground},
	{"root", "Root", XtRBoolean, sizeof(Boolean),
		 offset(root), XtRImmediate, (XtPointer)False},
	{"incinf", "IncInf", XtRBoolean, sizeof(Boolean),
		 offset(incinf), XtRImmediate, (XtPointer)False},
	{"num", "Num", XtRInt, sizeof(int),
		 offset(moduledatanum), XtRImmediate, (XtPointer)1},
};
#undef offset

static XtActionsRec actions[] =
{
	{"callback", Callback}
};

static char     translations[] = "";

XMaidClassRec xmaidClassRec = {
	{
		/* superclass         */ (WidgetClass) (&sessionShellClassRec),
		 /* class_name         */ "XMaid",
	     /* size               */ sizeof(XMaidRec),
		 /* class_initialize   */ ClassInitialize,
		 /* class_part_initialize */ NULL,
  	     /* Class init'ed      */ FALSE,
	     /* initialize         */ Initialize,
		 /* initialize_hook    */ NULL,
		 /* realize            */ Realize,
		 /* actions            */ actions,
		 /* num_actions        */ XtNumber(actions),
	     /* resources          */ resources,
		 /* resource_count     */ XtNumber(resources),
		 /* xrm_class          */ NULLQUARK,
		 /* compress_motion    */ FALSE,
		 /* compress_exposure  */ FALSE,
		 /* compress_enterleave */ FALSE,
		 /* visible_interest   */ FALSE,
		 /* destroy            */ Destroy,
		 /* resize             */ NULL,
		 /* expose             */ NULL,
		 /* set_values         */ SetValues,
		 /* set_values_hook    */ NULL,
		 /* set_values_almost  */ XtInheritSetValuesAlmost,
		 /* get_values_hook    */ NULL,
		 /* accept_focus       */ NULL,
		 /* intrinsics version */ XtVersion,
		 /* callback offsets   */ NULL,
		 /* tm_table		    */ translations,
		 /* query_geometry	    */ NULL,
		 /* display_accelerator */ NULL,
		 /* extension	        */ NULL
	},{
		/* geometry_manager   */    XtInheritGeometryManager,
		/* change_managed     */    XtInheritChangeManaged,
        /* insert_child   */    XtInheritInsertChild,
        /* delete_child   */    XtInheritDeleteChild,
        /* extension      */    NULL
    },{
        /* extension      */    NULL
    },{
        /* extension      */    NULL
    },{
        /* extension      */    NULL
    },{
        /* extension      */    NULL
    },{
		/* extension 	  */	NULL
    },{
        /* extension      */    NULL
    },{
		/* extension 	  */	NULL
	}
};

WidgetClass xmaidWidgetClass = (WidgetClass)&xmaidClassRec;

#define WIDTH			(xmw->core.width)
#define HEIGHT			(xmw->core.height)

#define ROOT			(xmw->xmaid.root)
#define ROOTGC			(xmw->xmaid.root_gc)

#define MODULEPATH		(xmw->xmaid.modulepath)
#define DATAPATH		(xmw->xmaid.datapath)

#define TIMERID			(xmw->xmaid.timer_id)
#define INPUTID			(xmw->xmaid.input_id)
#define MSGINFO			(xmw->xmaid.msginfo)
#define MSGINFOLIST		(xmw->xmaid.msginfolist)
#define MSGPATH			(xmw->xmaid.msgpath)
#define MOVEID 			(xmw->xmaid.move_id)

#define ANIMLIST		(xmw->xmaid.animlist)
#define MODULENAME		(xmw->xmaid.modulename)
#define MODULE			(xmw->xmaid.module)
#define MODULEINFO		(xmw->xmaid.module->info)
#define MODULEDATA		(xmw->xmaid.moduledatas)
#define MODULEDATANUM	(xmw->xmaid.moduledatanum)
#define MODULELEVEL 	(xmw->xmaid.modulelevel)

#define ANIM			(xmw->xmaid.moduledatas->anim)
#define XPOS			(xmw->xmaid.moduledatas->x)
#define YPOS			(xmw->xmaid.moduledatas->y)
#define XPOSBAK			(xmw->xmaid.moduledatas->xbak)
#define YPOSBAK			(xmw->xmaid.moduledatas->ybak)
#define ROOTBAK			(xmw->xmaid.moduledatas->root_bak)
#define ROOTWORK		(xmw->xmaid.moduledatas->root_work)

#define CURSORNAME		(xmw->xmaid.cursor_name)
#define CURSOR			(xmw->xmaid.cursor)
#define POINTERFG		(xmw->xmaid.pointer_fg)
#define POINTERBG		(xmw->xmaid.pointer_bg)

#define MENUX			(xmw->xmaid.menu_x)
#define MENUY			(xmw->xmaid.menu_y)
#define MENUDX			(xmw->xmaid.menu_dx)
#define MENUDY			(xmw->xmaid.menu_dy)

#define OLDDATAPATH		(old->xmaid.datapath)
#define OLDMODULEPATH	(old->xmaid.modulepath)

#define OLDMSGPATH		(old->xmaid.msgpath)

#define OLDMODULENAME	(old->xmaid.modulename)
#define OLDMODULE		(old->xmaid.module)
#define OLDCURSORNAME	(old->xmaid.cursor_name)
#define OLDCURSOR		(old->xmaid.cursor)
#define OLDPOINTERFG	(old->xmaid.pointer_fg)
#define OLDPOINTERBG	(old->xmaid.pointer_bg)

static void
ClassInitialize(void)
{
    static XtConvertArgRec convertArg[] = {
        {XtWidgetBaseOffset, (XtPointer)XtOffsetOf(WidgetRec, core.screen),
			 sizeof(Screen *)},
        {XtResourceString, (XtPointer) XtNpointerColor, sizeof(Pixel)},
        {XtResourceString, (XtPointer) XtNpointerColorBackground, 
			 sizeof(Pixel)},
        {XtWidgetBaseOffset, (XtPointer) XtOffsetOf(WidgetRec, core.colormap),
			 sizeof(Colormap)}
    };
    XtSetTypeConverter(XtRString, XtRColorCursor, XmuCvtStringToColorCursor,
					   convertArg, XtNumber(convertArg), 
					   XtCacheByDisplay, (XtDestructor)NULL);
}

static void
Initialize(Widget req, Widget new, ArgList args, Cardinal * num)
{
	XMaidWidget xmw = (XMaidWidget)new;

    TIMERID = 0;
	MOVEID  = 0;
	ANIMLIST = NULL;

	MSGINFO = xmaidmsginfo_new(xmw);
	MSGINFOLIST = NULL;
	
	MENUX = MENUY = 0;
	MENUDX = MENUDY = 0;

	MODULE        = &xmaid_default_module;
	MODULEDATA    = &xmaid_default_moduledata;
	MODULELEVEL	  = NULL;

	if (!ROOT)
		MODULEDATANUM = 1;

	if (MODULENAME) {
		MODULE     = xmaid_module_open(xmw, MODULENAME);
		MODULEDATA = xmaid_moduledata_open(xmw, MODULE, MODULEDATANUM);
	}

	XtVaSetValues(new,
				  XtNoverrideRedirect, MODULEINFO->override, 
				  XtNsaveUnder, MODULEINFO->override, NULL);

	ROOTGC  = DefaultGC(XtDisplay((Widget)xmw),
						XScreenNumberOfScreen(XtScreen((Widget)xmw)));
}

static void
Destroy(Widget widget)
{
	XMaidWidget xmw = (XMaidWidget)widget;

	if (TIMERID)
		XtRemoveTimeOut(TIMERID);
	if (MOVEID)
		XtRemoveTimeOut(MOVEID);

	xmaidmsginfo_delete(MSGINFO);
	xmaidmsginfolist_delete(MSGINFOLIST);

	if (ROOT) {
		if (ROOTBAK) {
			XCopyArea(XtDisplay((Widget)xmw), ROOTBAK, ANIM->win, 
					  ROOTGC, 0, 0, WIDTH, HEIGHT, XPOSBAK, YPOSBAK);
			XFreePixmap(XtDisplay((Widget)xmw), ROOTBAK);
		}
		if (ROOTWORK) {
			XFreePixmap(XtDisplay((Widget)xmw), ROOTWORK);
		}
	}

	if (ANIMLIST)
		timerlist_delete(ANIMLIST);

	if (MODULEDATA != &xmaid_default_moduledata) {
		int i;
		for (i=0;i<MODULEDATANUM;i++)
			xmaid_moduledata_destroy(xmw, MODULE, MODULEDATA+i);
	}
	if (MODULE != &xmaid_default_module)
		xmaid_module_destroy(xmw, MODULE);

	if (MODULELEVEL)
		free(MODULELEVEL);
}

static 
int xmaidmoddat_compare(const void *a, const void *b)
{
	XMaidModuleData **a1 = (XMaidModuleData **)a;
	XMaidModuleData **b1 = (XMaidModuleData **)b;
	return (*b1)->y - (*a1)->y;
}

static void
Realize(Widget widget, XtValueMask *value_mask, XSetWindowAttributes *attr)
{
	XMaidWidget xmw = (XMaidWidget)widget;
	XtRealizeProc realize;
	int i;

	MOVEID = 0;

	xmaid_module_init(xmw, MODULE);
	xmaid_moduledata_init(xmw, MODULE, MODULEDATA);
	for (i=1;i<MODULEDATANUM;i++) {
		xmaid_moduledata_dup(xmw, MODULE, MODULEDATA, MODULEDATA+i);
	}

	if (ANIM) {
		WIDTH  = ANIM->w;
		HEIGHT = ANIM->h;
	}

	ConvertCursor(widget);
	if ((attr->cursor = xmw->xmaid.cursor) != None)
		*value_mask |= CWCursor;
	realize = xmaidWidgetClass->core_class.superclass->core_class.realize;
	(*realize)(widget, value_mask, attr);

	if (!ROOT) {
		animate_attach_win(ANIM, XtDisplay((Widget)xmw), 
						   XtWindow((Widget)xmw));
		animate_realize(ANIM);
		TimerHandler(widget, NULL);
		if (MODULEINFO->map == False) {
			XtSetMappedWhenManaged((Widget)xmw, False);
			if (MODULEINFO->update_timing == 0)
				MoveHandler(xmw, NULL);
		}
		XtAddEventHandler(widget, StructureNotifyMask, FALSE, 
						  StartCallback, NULL);
	} else {
		ANIMLIST = timer_new();
		if ((MODULELEVEL = malloc(sizeof(*MODULELEVEL) * MODULEDATANUM))
			== NULL) {
			perror("xmaid:realize:");
			exit(1);
		}
		for (i=0;i<MODULEDATANUM;i++)
			MODULELEVEL[i] = MODULEDATA+i;

		qsort(MODULELEVEL, MODULEDATANUM, sizeof(*MODULELEVEL), 
			  xmaidmoddat_compare);

		/* ͥ٤ι⤤Τ饨ȥ */
		for (i=0;i<MODULEDATANUM;i++) {
			XMaidModuleData *p = MODULELEVEL[i];
			p->level = i;
			animate_attach_win(p->anim, XtDisplay((Widget)xmw), 
							   RootWindowOfScreen(XtScreen((Widget)xmw)));
			animate_realize(p->anim);
#if 0
			{
				XWindowAttributes attr;
				Display *dpy = XtDisplay(xmw);
				Window root = RootWindowOfScreen(XtScreen(xmw));
				XGetWindowAttributes(dpy, root, &attr);
				fprintf(stderr, "window attr:%x %lx %lx\n", 
						root, attr.your_event_mask, ExposureMask);
			}
#endif
			p->prev_anim = p->anim;
			timer_entry(ANIMLIST,p);
		}

		XtSetMappedWhenManaged((Widget)xmw, False);
		TimerHandler2(widget, NULL);
	}

}

/* from Xaw Simple Widget */

/*  Function Name: ConvertCursor
 *  Description: Converts a name to a new cursor.
 *  Arguments: w - the simple widget.
 *  Returns: none.
 */

static void
ConvertCursor(Widget w)
{
    XMaidWidget xmw = (XMaidWidget) w;
    XrmValue from, to;
    Cursor cursor;

    if (CURSORNAME == NULL)
		return;

    from.addr = (XPointer)CURSORNAME;
    from.size = strlen((char *) from.addr) + 1;
    to.size = sizeof(Cursor);
    to.addr = (XPointer)&cursor;

    if (XtConvertAndStore(w, XtRString, &from, XtRColorCursor, &to)) {
		if (cursor !=  None) 
			CURSOR = cursor;
    } else{
		XtAppErrorMsg(XtWidgetToApplicationContext(w),
					  "convertFailed","ConvertCursor","XMaidWidgetError",
					  "Simple: ConvertCursor failed.",
					  (String *)NULL, (Cardinal *)NULL);
    }
}

static Boolean
SetValues(Widget cur, Widget req, Widget new, ArgList args, Cardinal *num_args)
{
	XMaidWidget old = (XMaidWidget)cur;
	XMaidWidget xmw = (XMaidWidget)new;
	Boolean redisplay = FALSE;
	Boolean new_cursor = FALSE;

	if (OLDMODULEPATH != MODULEPATH) {
		if (OLDMODULEPATH)
			free(OLDMODULEPATH);
		if (MODULEPATH)
			MODULEPATH = strdup(MODULEPATH);
	}

	if (OLDMSGPATH != MSGPATH) {
		if (OLDMSGPATH)
			free(OLDMSGPATH);
		if (MSGPATH)
			MSGPATH = strdup(MSGPATH);
	}

	if (OLDDATAPATH != DATAPATH) {
		if (OLDDATAPATH)
			free(OLDDATAPATH);
		if (DATAPATH)
			DATAPATH = strdup(DATAPATH);
	}

	/* for cursor */
	if (OLDCURSOR != CURSOR)
		new_cursor = TRUE;
	if ((OLDPOINTERFG != POINTERFG) ||
		(OLDPOINTERBG != POINTERBG) ||
		(OLDCURSORNAME != CURSORNAME)) {
		ConvertCursor(new);
		new_cursor = TRUE;
	}
	if (new_cursor && XtIsRealized(new))
		XDefineCursor(XtDisplay(new), XtWindow(new), CURSOR);

	return redisplay;
}

/* ŤŤʻ롼 */
#if 0
/*--------------------------------------------------------------------*/
	if ((dx = abs(XPOS2 - XPOSBAK2)) < WIDTH && 
		(dy = abs(YPOS2 - YPOSBAK2)) < HEIGHT) {
		if (XPOS2 < XPOSBAK2) {
			if (YPOS2 < YPOSBAK2) {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2, YPOS2, WIDTH, HEIGHT, 0, 0);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  0, 0, WIDTH-dx, HEIGHT-dy, dx, dy);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  0, HEIGHT-dy, WIDTH-dx, dy, XPOSBAK2, YPOSBAK2+HEIGHT-dy);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  WIDTH-dx, 0, dx, HEIGHT, XPOSBAK2+WIDTH-dx, YPOSBAK2);
			} else if (YPOS2 > YPOSBAK2) {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2, YPOS2, WIDTH, HEIGHT, 0, 0);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  0, dy, WIDTH-dx, HEIGHT-dy, dx,0);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  0, 0, WIDTH-dx, dy, XPOSBAK2, YPOSBAK2);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  WIDTH-dx, 0, dx, HEIGHT, XPOSBAK2+WIDTH-dx, YPOSBAK2);
			} else {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2, YPOS2, dx, HEIGHT, 0, 0);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  0, 0, WIDTH-dx, HEIGHT, dx, 0);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  WIDTH-dx, 0, dx, HEIGHT, XPOSBAK2+WIDTH-dx, YPOSBAK2);
			}
		} else if (XPOS2 > XPOSBAK2) {
			if (YPOS2 < YPOSBAK2) {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2, YPOS2, WIDTH, HEIGHT, 0, 0);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  dx, 0, WIDTH-dx, HEIGHT-dy, 0, dy);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  0, 0, dx, HEIGHT, XPOSBAK2, YPOSBAK2);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  dx, HEIGHT-dy, WIDTH-dx, dy, XPOSBAK2+dx, YPOSBAK2+HEIGHT-dy);
			} else if (YPOS2 > YPOSBAK2) {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2, YPOS2, WIDTH, HEIGHT, 0, 0);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  dx, dy, WIDTH-dx, HEIGHT-dy, 0, 0);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  dx, 0, WIDTH-dx, dy, XPOSBAK2+dx, YPOSBAK2);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  0, 0, dx, HEIGHT, XPOSBAK2, YPOSBAK2);
			} else {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2+WIDTH-dx, YPOS2, dx, HEIGHT, WIDTH-dx, 0);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  dx, 0, WIDTH-dx, HEIGHT, 0, 0);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  0, 0, dx, HEIGHT, XPOSBAK2, YPOSBAK2);
			}
		} else {
			if (YPOS2 < YPOSBAK2) {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2, YPOS2, WIDTH, dy, 0, 0);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  0, 0, WIDTH, HEIGHT-dy, 0, dy);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  0, HEIGHT-dy, WIDTH, dy, XPOSBAK2, YPOSBAK2+HEIGHT-dy);
			} else if (YPOS2 > YPOSBAK2) {
				XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
						  XPOS2, YPOS2+HEIGHT-dy, WIDTH, dy, 0, HEIGHT-dy);
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  0, dy, WIDTH, HEIGHT-dy, 0, 0);
				XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
						  0, 0, WIDTH, dy, XPOSBAK2, YPOSBAK2);
			} else {
				XCopyArea(dpy, ROOTBAK2, ROOTWORK2, ROOTGC,
						  0, 0, WIDTH, HEIGHT, 0, 0);
			}
		}
	} else {
		XCopyArea(dpy, win, ROOTWORK2, ROOTGC,
				  XPOS2, YPOS2, WIDTH, HEIGHT, 0, 0);
		XCopyArea(dpy, ROOTBAK2, win, ROOTGC,
				  0, 0, WIDTH, HEIGHT, XPOSBAK2, YPOSBAK2);
	}
/*--------------------------------------------------------------------*/
#endif

/* Ťʻ 1 ŤʤäƤʬ */
static void
kasane(XMaidWidget xmw, Drawable a, int x1, int y1, Drawable b, int x2, int y2) 
{
	int dx, dy;
	Display *dpy = XtDisplay((Widget)xmw);

	if ((dx = abs(x1 - x2)) < WIDTH && 
		(dy = abs(y1 - y2)) < HEIGHT) {
		if (x2 < x1) {
			if (y2 < y1)
				XCopyArea(dpy, a, b, ROOTGC,
						  0, 0, WIDTH-dx, HEIGHT-dy, dx, dy);
			else
				XCopyArea(dpy, a, b, ROOTGC,
						  0, dy, WIDTH-dx, HEIGHT-dy, dx,0);
		} else {
			if (y2 < y1)
				XCopyArea(dpy, a, b, ROOTGC,
						  dx, 0, WIDTH-dx, HEIGHT-dy, 0, dy);
			else
				XCopyArea(dpy, a, b, ROOTGC,
						  dx, dy, WIDTH-dx, HEIGHT-dy, 0, 0);
		}
	}
}

/* Ťͤ碌2 ŤʤäƤʤʬ */
static void
kasane2(XMaidWidget xmw, Drawable a, int x1, int y1, Drawable b, int x2, int y2) 
{
	int dx, dy;
	Display *dpy = XtDisplay((Widget)xmw);

	if ((dx = abs(x1 - x2)) < WIDTH && 
		(dy = abs(y1 - y2)) < HEIGHT) {
		if (x2< x1) {
			if (y2 < y1) {
				XCopyArea(dpy, a, b, ROOTGC,
						  0, HEIGHT-dy, WIDTH-dx, dy, x1, y1+HEIGHT-dy);
				XCopyArea(dpy, a, b, ROOTGC,
						  WIDTH-dx, 0, dx, HEIGHT, x1+WIDTH-dx, y1);
			} else {
				XCopyArea(dpy, a, b, ROOTGC,
						  0, 0, WIDTH-dx, dy, x1, y1);
				XCopyArea(dpy, a, b, ROOTGC,
						  WIDTH-dx, 0, dx, HEIGHT, x1+WIDTH-dx, y1);
			}
		} else
			if (y2 < y1) {
				XCopyArea(dpy, a, b, ROOTGC,
						  0, 0, dx, HEIGHT, x1, y1);
				XCopyArea(dpy, a, b, ROOTGC,
						  dx, HEIGHT-dy, WIDTH-dx, dy, x1+dx, y1+HEIGHT-dy);
			} else {
				XCopyArea(dpy, a, b, ROOTGC,
						  dx, 0, WIDTH-dx, dy, x1+dx, y1);
				XCopyArea(dpy, a, b, ROOTGC,
						  0, 0, dx, HEIGHT, x1, y1);
			}
	} else {
		XCopyArea(dpy, a, b, ROOTGC,
				  0, 0, WIDTH, HEIGHT, x1, y1);
	}
}

/* 񤭹߽ ŤʤäƤʬΤ */
static void
draw(XMaidWidget xmw, Drawable a, Drawable mask, int x1, int y1, Drawable b, int x2, int y2)
{
	int offx = x1-x2;
	int offy = y1-y2;
	int dx,dy;

	Display *dpy = XtDisplay((Widget)xmw);

	if ((dx = abs(offx)) < WIDTH && 
		(dy = abs(offy)) < HEIGHT) {

		XSetClipMask(dpy, ROOTGC, mask);
		XSetClipOrigin(dpy, ROOTGC, offx, offy);

		if (x2 < x1) {
			if (y2 < y1)
				XCopyArea(dpy, a, b, ROOTGC,
						  0, 0, WIDTH-dx, HEIGHT-dy, dx, dy);
			else
				XCopyArea(dpy, a, b, ROOTGC,
						  0, dy, WIDTH-dx, HEIGHT-dy, dx,0);
		} else {
			if (y2 < y1)
				XCopyArea(dpy, a, b, ROOTGC,
						  dx, 0, WIDTH-dx, HEIGHT-dy, 0, dy);
			else
				XCopyArea(dpy, a, b, ROOTGC,
						  dx, dy, WIDTH-dx, HEIGHT-dy, 0, 0);
		}

		XSetClipMask(dpy, ROOTGC, None);
		XSetClipOrigin(dpy, ROOTGC, 0, 0);

	}
}

#define ANIM2			(data->anim)
#define PREVANIM2		(data->prev_anim)
#define XPOS2			(data->x)
#define YPOS2			(data->y)
#define XPOSBAK2		(data->xbak)
#define YPOSBAK2		(data->ybak)
#define ROOTBAK2		(data->root_bak)
#define ROOTWORK2		(data->root_work)

void
XMaidExposeRoot(XMaidWidget xmw, XEvent *ev)
{
	int x = ev->xexpose.x;
	int y = ev->xexpose.y;
	int w = ev->xexpose.width;
	int h = ev->xexpose.height;

	XClearArea(ev->xexpose.display, ev->xexpose.window,
			   x, y, w, h, False);

	fprintf(stderr, "expose root!!!1\n");

	if (ROOT) {
		int i;
		XMaidModuleData *p;
		XMaidModuleData *data = MODULEDATA;
		for (i=0;i<MODULEDATANUM;i++,data++) {
			if (ROOTBAK2 != None &&
				(XPOS2 < x+w && XPOS2+WIDTH > x) &&
				(YPOS2 < y+h && YPOS2+HEIGHT > y)) {
				XCopyArea(ANIM2->dpy, ANIM2->win, ROOTBAK2, 
						  ROOTGC, x, y, w, h, x-XPOS2, y-YPOS2);
			}
		}
	}
}

static int
draw_animate(XMaidWidget xmw, XMaidModuleData *data)
{
	int i;
	XMaidModuleData *p;
	Animate *anim = ANIM2;
	Display *dpy = anim->dpy;
	Window   win = anim->win;
	int ret;
	int dx, dy;
	ret = animate_go(anim);

	if (xmw->xmaid.incinf)
		XSetSubwindowMode(dpy, ROOTGC, IncludeInferiors);

	{
		XEvent ev;
	}

	if (ROOTBAK2 == None) {
		ROOTBAK2 = XCreatePixmap(dpy, win, WIDTH, HEIGHT, anim->depth);
		ROOTWORK2 = XCreatePixmap(dpy, win, WIDTH, HEIGHT, anim->depth);
		XCopyArea(dpy, win, ROOTBAK2, ROOTGC, XPOS2, YPOS2, WIDTH, HEIGHT, 0, 0);
		for (i=0;i<data->level;i++) {
			p = MODULELEVEL[i];
			kasane(xmw, p->root_bak, p->x, p->y, data->root_bak, data->x, data->y);
		}
		XPOSBAK2 = XPOS2;
		YPOSBAK2 = YPOS2;
	} 

	/* ư⥸塼 */
	if (MODULEINFO->update_timing == 1 && MODULEINFO->move)
		MODULEINFO->move(xmw, data);

	/* ɽΰ */
	/* copy from rootwin */
	XCopyArea(dpy, win, ROOTWORK2, ROOTGC, XPOS2, YPOS2, WIDTH, HEIGHT, 0, 0);

	/* copy from upper moduledata's back */
	for (i=0;i<data->level;i++) {
		p = MODULELEVEL[i];
		kasane(xmw,p->root_bak, p->x, p->y, data->root_work, data->x, data->y);
	}
	/* copy from my back */
	kasane(xmw, data->root_bak, data->xbak, data->ybak, data->root_work, data->x, data->y);

	/* ΰõ */
	/* 缡̤ȿ */
	for (i=data->level-1;i>=0;i--) {
		p = MODULELEVEL[i];
		kasane(xmw, data->root_bak, data->xbak, data->ybak, p->root_bak, p->x, p->y);
		draw(xmw, p->prev_anim->pixmap, p->prev_anim->mask, p->x, p->y , data->root_bak, data->xbak, data->ybak);
	}
#if 1
	/* ǽõ */
	kasane2(xmw, data->root_bak, data->xbak, data->ybak, win, data->x, data->y);
#endif

	/* y ɸѹ --> ͥѹ */
	if (YPOS2 != YPOSBAK2) {
		int l = data->level;
		if (YPOSBAK2 < YPOS2) {		/* ͥ̾徺 */
			while (l > 0) {
				if (MODULELEVEL[l-1]->y < YPOS2) {
					MODULELEVEL[l] = MODULELEVEL[l-1];
					MODULELEVEL[l]->level = l;
					l--;
				} else
					break;
			}
			MODULELEVEL[l] = data;
			data->level = l;
		} else {					/* ̸ͥ */
			while (l < MODULEDATANUM-1) {
				if (MODULELEVEL[l+1]->y > YPOS2) {
					MODULELEVEL[l] = MODULELEVEL[l+1];
					MODULELEVEL[l]->level = l;
					l++;
				} else
					break;
			}
			MODULELEVEL[l] = data;
			data->level = l;
		}
	}

	/*  */
	/* ư⥸塼뤽2 (root_work ؽ񤭹) */
	if (MODULEINFO->update_timing == 1 && MODULEINFO->after_move)
		MODULEINFO->after_move(xmw, data);
	/* õ (work->bak) */
	XCopyArea(dpy, ROOTWORK2, ROOTBAK2, ROOTGC,  0, 0, WIDTH, HEIGHT, 0, 0);
	/* ʬΥǡ */
	draw(xmw, anim->pixmap, anim->mask, data->x, data->y, data->root_work, data->x, data->y);
#if 1
	/* 缡̤ȿ */
	for (i=data->level-1;i>=0;i--) {
		p = MODULELEVEL[i];
		kasane(xmw, data->root_work, data->x, data->y, p->root_bak, p->x, p->y);
		draw(xmw, p->prev_anim->pixmap, p->prev_anim->mask, p->x, p->y, data->root_work, data->x, data->y);
	}
#endif

	/* ǽ */
	XCopyArea(dpy, ROOTWORK2, win, ROOTGC, 0, 0, WIDTH, HEIGHT, XPOS2, YPOS2);
	XSync(dpy, True);

	/* new animateion pattern */
	if (anim != ANIM2) {
		animate_attach_win(ANIM2, XtDisplay((Widget)xmw), 
						   RootWindowOfScreen(XtScreen((Widget)xmw)));
		animate_realize(ANIM2);	
		if (!ret) ret = 1;
	}
	XPOSBAK2 = XPOS2;
	YPOSBAK2 = YPOS2;
	PREVANIM2 = anim;

	XSetSubwindowMode(dpy, ROOTGC, None);

	return ret;
}


static u_int
timer_go(XMaidWidget xmw)
{
	int ret;
	u_int interval;
	XMaidModuleData *data;
	TimerDat *p;

	if ((p = ANIMLIST->next) != NULL) {
	retry:
		data = p->data;
		interval = data->anim->point->interval;
#if 1
		ret = draw_animate(xmw, data);
#else
		ret = 1;
#endif
		ANIMLIST->next = p->next;
		if (ret) {
			if (interval == 0)
				interval = 10;	/* XXX */
			timer_reentry(ANIMLIST, p->data, p, interval);
		} else
			timer_delete(p);
		if ((p = ANIMLIST->next)) {
			if (ANIMLIST->next->interval == 0)
				goto retry;
			return ANIMLIST->next->interval;
		}
	}
	return 0;
}

/* timer handler for animated gif */
static void
TimerHandler2(XtPointer cl, XtIntervalId *id)
{
	XMaidWidget xmw = cl;
	u_int interval;
	if ((interval = timer_go(xmw)) > 0)
		TIMERID = XtAppAddTimeOut(XtWidgetToApplicationContext(cl),
								  interval,  TimerHandler2, cl);
	else
		TIMERID = 0;
}

/* timer handler for animated gif */
static void
TimerHandler(XtPointer cl, XtIntervalId *id)
{
	XMaidWidget xmw = cl;
	Animate *anim = ANIM;
	Display *dpy = anim->dpy;
	Window   win = anim->win;

	int ret;
	u_int interval;

	interval = anim->point->interval;
	ret = animate_go(anim);

	/* ư⥸塼 */
	if (MODULEINFO->update_timing == 1 && MODULEINFO->move) {
		MODULEINFO->move(xmw, MODULEDATA);
		if (XPOS != XPOSBAK || YPOS != YPOSBAK)
			XtMoveWidget((Widget)xmw, XPOS, YPOS);
		XPOSBAK = XPOS;
		YPOSBAK = YPOS;
	}

	animate_set_shape(anim);
	XSetWindowBackgroundPixmap(dpy, win, anim->pixmap);
	XClearWindow(dpy, win);
	XSync(dpy, None);

	/* new animateion pattern */
	if (anim != ANIM) {
		animate_attach_win(ANIM, XtDisplay((Widget)xmw), 
						   (ROOT)?
						   RootWindowOfScreen(XtScreen((Widget)xmw)):
						   XtWindow((Widget)xmw));
		animate_realize(ANIM);	
		if (!ret) ret = 1;
	}

	if (ret)
		TIMERID = XtAppAddTimeOut(XtWidgetToApplicationContext(cl),
								  interval,  TimerHandler, cl);
	else {
		TIMERID = 0;
		if (MOVEID == 0 && MODULEINFO->update_timing == 0)
			MoveHandler(xmw, NULL);
	}
}


static void
MoveHandler(XtPointer cl, XtIntervalId *id)
{
	XMaidWidget xmw = cl;
	int ret;
	if (MODULEINFO->move) {
		Animate *anim = ANIM;
		ret = MODULEINFO->move(xmw, MODULEDATA);
		if (XPOS != XPOSBAK || YPOS != YPOSBAK)
			XtMoveWidget((Widget)xmw, XPOS, YPOS);
		XPOSBAK = XPOS;
		YPOSBAK = YPOS;

		/* new animateion pattern */
		if (anim != ANIM) {
			animate_attach_win(ANIM, XtDisplay((Widget)xmw), 
							   (ROOT)?
							   RootWindowOfScreen(XtScreen((Widget)xmw)):
							   XtWindow((Widget)xmw));
			animate_realize(ANIM);	
			if (TIMERID == 0)
				TimerHandler(xmw, NULL);
		}

		if (ret > 0)
			MOVEID = XtAppAddTimeOut(XtWidgetToApplicationContext(cl),
									 ret,  MoveHandler, cl); 
		else
			MOVEID = 0;
	} else
		MOVEID = 0;
}

static void
StartCallback(Widget widget, XtPointer client_data, XEvent *ev, Boolean *cont)
{
	XMaidWidget xmw = (XMaidWidget)widget;

	if (ev->type == MapNotify) {
		if (MOVEID == 0) {
			MOVEID = XtAppAddTimeOut(XtWidgetToApplicationContext(widget),
									 10,  MoveHandler, widget);
		}
		XtCallCallbackList(widget, xmw->xmaid.startup_callbacks, 
						   (XtPointer)MSGINFO);
		XtRemoveEventHandler(widget, StructureNotifyMask, FALSE, 
							 StartCallback, client_data);
	}
}

static void
Callback(Widget w, XEvent *ev, String *p, Cardinal *n)
{
	XMaidWidget xmw = (XMaidWidget)w;
	XtCallCallbackList(w, xmw->xmaid.callbacks, (XtPointer)MSGINFO);
}

char *
XMaidDataPath(XMaidWidget xmw, char *name)
{
	static char buf[1024];	/* XXX */
	sprintf(buf, "%s/%s", DATAPATH, name);
	return buf;
}

void
XMaidStopMove(XMaidWidget xmw)
{
	if (TIMERID)
		XtRemoveTimeOut(TIMERID);
	TIMERID = 0;
	if (MOVEID)
		XtRemoveTimeOut(MOVEID);
	MOVEID = 0;
}

void
XMaidStartMove(XMaidWidget xmw)
{
	if (MOVEID == 0 && MODULEINFO->update_timing == 0)
		MoveHandler(xmw, NULL);
	if (TIMERID == 0 && ANIM)
		TimerHandler(xmw, NULL);
}

void
XMaidAddMsgInfo(XMaidWidget xmw, XMaidMsgInfo *msg)
{
	msg->next = MSGINFOLIST;
	MSGINFOLIST = msg;
}

void
XMaidDeleteMsgInfo(XMaidWidget xmw, XMaidMsgInfo *msg)
{
	XMaidMsgInfo **p = &MSGINFOLIST, *q;
	q = MSGINFOLIST;
	while (q != NULL) {
		if (msg == q) {
			*p = q->next;
			break;
		}
		p = &q->next;
		q = q->next;
	}
}
