#ifndef __AnimCoreP_h
#define __AnimCoreP_h

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/ShellP.h>
#include "XMaid.h"

#include "timer.h"

#include "animate.h"
#include "xmaidmsg.h"
#include "xmaidmod.h"

typedef struct{
	XtPointer extension;
} XMaidClassPart;

typedef struct _XMaidClassRec{
	CoreClassPart core_class;
    CompositeClassPart composite_class;
    ShellClassPart  shell_class;
    WMShellClassPart   wm_shell_class;
    VendorShellClassPart vendor_shell_class;
    TopLevelShellClassPart top_level_shell_class;
	ApplicationShellClassPart application_shell_class;
    SessionShellClassPart session_shell_class;
	XMaidClassPart xmaid_class;
} XMaidClassRec;

extern XMaidClassRec xmaidClassRec;

typedef struct _XMaidPart {

	XtCallbackList startup_callbacks;
	XtCallbackList callbacks;
	XtIntervalId timer_id;
	XtIntervalId  move_id;

	XMaidMsgInfo *msginfo;
	XMaidMsgInfoList msginfolist;

	char *modulename;
	XMaidModule  	 *module;
	XMaidModuleData  *moduledatas;
	int moduledatanum;
	XMaidModuleData **modulelevel;

	char *modulepath;
	char *msgpath;
	char *datapath;

	/* from simple widget */
	Pixel  pointer_fg, pointer_bg; /* Pointer colors. */
	Cursor cursor;
	String cursor_name;

	/* menu base position */
	int menu_x;
	int menu_y;
	int menu_dx;
	int menu_dy;

	Boolean root;
	Boolean incinf;
	GC	root_gc;

	TimerDatList animlist;

} XMaidPart;

typedef struct _XMaidRec {
	CorePart			 core;
	CompositePart   	 composite;
    ShellPart   		 shell;
    WMShellPart 		 wm;
    VendorShellPart 	 vendor;
    TopLevelShellPart 	 topLevel;
	ApplicationShellPart application;
    SessionShellPart 	 session;
	XMaidPart 			 xmaid;
} XMaidRec;

/* private */
void XMaidStopMove(XMaidWidget w);
void XMaidStartMove(XMaidWidget w);

void XMaidAddMsgInfo(XMaidWidget xmw, XMaidMsgInfo *msg);
void XMaidDeleteMsgInfo(XMaidWidget xmw, XMaidMsgInfo *msg);

#endif
