#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "child.h"

typedef struct ChildProcessInfo {
	struct ChildProcessInfo *next;
	pid_t pid;
	char *name;
} ChildProcessInfo, *ChildProcessInfoList;

static ChildProcessInfo *
childprocessinfo_new(void)
{
	ChildProcessInfo *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("childprocessinfo_new");
		exit(1);
	}
	p->next = NULL;
	p->pid = 0;
	p->name = NULL;
	return p;
}

static void
childprocessinfo_delete(ChildProcessInfo *p)
{
	if (p) {
		if (p->name)
			free(p->name);
		free(p);
	}
}

static void
childprocessinfo_add(ChildProcessInfoList *l, pid_t pid, const char *name)
{
	ChildProcessInfo *c = childprocessinfo_new();
	c->pid = pid;
	c->name = strdup(name);
	c->next = *l;
	*l = c;
}

static void
childprocessinfo_clear(ChildProcessInfoList *l, pid_t pid)
{
	ChildProcessInfo *p;
	p = *l;
	while (p != NULL) {
		if (p->pid == pid) {
			*l = p->next;
			childprocessinfo_delete(p);
			break;
		}
		l = &(p->next);
		p = p->next;
	}
}

static ChildProcessInfoList childlist = NULL;

void
childinfo_entry(pid_t pid, const char *p)
{
	childprocessinfo_add(&childlist, pid, p);
}

void
childinfo_term(int dummy)
{
	pid_t pid;
	pid = wait(NULL);
	childprocessinfo_clear(&childlist, pid);
}

void
childinfo_kill(pid_t pid)
{
	ChildProcessInfo *p = childlist;	
	while (p != NULL) {	
		if (p->pid == pid) {
			kill(pid, SIGTERM);
			break;
		}
	}
}

void
childinfo_show(FILE *fp)
{
	ChildProcessInfo *p = childlist;	

	if (p) {
		fprintf(fp, "<ul>");
		while (p != NULL) {	
			fprintf(fp, "<li><a href=\"xmaid:kill:%d\">[KILL]</a>:%s\n",
					p->pid, p->name);
			p = p->next;
		}
		fprintf(fp, "</ul>");
	}
}

/* XXX  this is bad function NO CHECK */
void
child_exec(char *name)
{
	pid_t pid;
	const char *p;
	
	if ((p = strrchr(name, '/')) == NULL)
		p = name;
	else
		p++;

	if ((pid = vfork()) < 0)	
		perror("child_exec:vfork");
	else if (pid == 0) {
		int i, n, fdin;

		if ((n = sysconf(_SC_OPEN_MAX)) < 0) {
			if (errno == 0)
				n = 256;
			else 
				perror("sysconf: _SC_OPEN_MAX");
		}
		for (i=2;i<n;i++)
			close(i);

		if ((fdin = open("/dev/null", O_RDONLY)) < 0) {
			perror("xmaidmsginfo_exec_file:open");
			exit(1);
		}
		close(0);
		dup(fdin);

		execlp(name, p, NULL);
		perror("xmaidmsginfo_exec:execlp");
		exit(1);
	}
	childinfo_entry(pid, p);
}
