#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <sys/types.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "xlocale.h"

#include "MsgShell.h"
#include "XMaid.h"

#include "xmaidmsg.h"
#include "xmaidmod.h"
#include "child.h"

#ifndef XMAIDSTARTUP
#define XMAIDSTARTUP "start.xhtml"
#endif

#ifndef XMAIDMAINMENU
#define XMAIDMAINMENU "menu.html"
#endif

static void
Destroy(Widget w, XEvent *e, String *p, Cardinal *n)
{
	XtDestroyWidget(w);
}

static String trans = "<Btn3Up>: destroy()";

static void 
start_popup(Widget w, XtPointer client_data, XtPointer call_data)
{
	XMaidMsgInfo *msg = call_data;
	if (!msg->inputflag) {
		xmaid_msginfo_open2((XMaidWidget)w, 600, 500,
							XMaidDataPath((XMaidWidget)w,XMAIDSTARTUP));
		XtOverrideTranslations(msg->msgbuf, XtParseTranslationTable(trans));
	}
}

static void
menu_popup(Widget w, XtPointer client_data, XtPointer call_data)
{
	XMaidMsgInfo *msg = call_data;
	if (!msg->inputflag) {
		xmaid_msginfo_open2((XMaidWidget)w, 600, 500,
							XMaidDataPath((XMaidWidget)w,XMAIDMAINMENU));
		XtOverrideTranslations(msg->msgbuf, XtParseTranslationTable(trans));
	}
}

/* command-line options */
static XrmOptionDescRec options[] = {
	{ "-modulepath", 	"." XtNmodulePath, 	XrmoptionSepArg, NULL },
	{ "-datapath", 		"." XtNdataPath,	XrmoptionSepArg, NULL },
	{ "-mod",			"." XtNmoduleName,	XrmoptionSepArg, NULL },
	{ "-module", 		"." XtNmoduleName,	XrmoptionSepArg, NULL },
	{ "-root", 			".root", 			XrmoptionNoArg, (XtPointer)"True"},
	{ "-incinf", 		".incinf", 			XrmoptionNoArg, (XtPointer)"True"},
	{ "-num", 			".num",  			XrmoptionSepArg, NULL },
};

/* actions */
static XtActionsRec actions[] = {
	{"destroy", Destroy},
};

/* translation */
String trans_top = "<Btn3Up>: callback()";

static char *event_names[] = {
	"", "", "KeyPress", "KeyRelease", "ButtonPress",
	"ButtonRelease", "MotionNotify", "EnterNotify",
	"LeaveNotify", "FocusIn", "FocusOut", "KeymapNotify",
	"Expose", "GraphicsExpose", "NoExpose", "VisibilityNotify",
	"CreateNotify", "DestroyNotify", "UnmapNotify", "MapNotify",
	"MapRequest", "ReparentNotify", "ConfigureNotify", "ConfigureRequest",
	"GravityNotify", "ResizeRequest", "CirculateNotify", "CirculateRequest",
	"PropertyNotify", "SelectionClear", "SelectionRequest",
	"SelectionNotify", "ColormapNotify", "ClientMessage", "MappingNotify"
};


int
main(int argc, char **argv)
{
	XtAppContext app;
	Widget top;
	Window root;

	srandom((unsigned)getpid());
    signal(SIGCHLD, childinfo_term);

	mysetlocale(0);

	top = XtVaOpenApplication(&app, "XMaid", options, XtNumber(options),
							  &argc, argv, NULL, xmaidWidgetClass, 
							  NULL);

	XtAppAddActions(app, actions, XtNumber(actions));
	XtOverrideTranslations(top, XtParseTranslationTable(trans_top));

	XtAddCallback(top, XtNstartupCallback, start_popup, NULL);
	XtAddCallback(top, XtNcallback, menu_popup, NULL);
	
#if 0
	root = RootWindowOfScreen(XtScreen(top));
	XSelectInput(XtDisplay(top), root, ExposureMask);
	fprintf(stderr, "Display:%p Screem:%pRootWindow: %x\n", 
			XtDisplay(top), XtScreen(top), root);
	{
		XWindowAttributes attr;
		Display *dpy = XtDisplay(top);
		Window root = RootWindowOfScreen(XtScreen(top));
		XGetWindowAttributes(dpy, root, &attr);
		fprintf(stderr, "start window attr:%x %lx %lx\n", 
				root, attr.your_event_mask, ExposureMask);
	}
#endif

	XtRealizeWidget(top);

	do {
		XEvent event;
		XtAppNextEvent(app, &event);
/*
		if (event.type == Expose && event.xany.window == root)
			XMaidExposeRoot((XMaidWidget)top, &event);
		fprintf(stderr, "type: %s window:%x\n", event_names[event.type],
				event.xany.window);
*/
		XtDispatchEvent(&event);
    } while (XtAppGetExitFlag(app) == FALSE);
	
	exit(0);
}
