#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "MsgShell.h"
#include "XMaidP.h"
#include "xmaidmsg.h"
#include "child.h"

#define DEFAULT_FILE "index.html"

#define MSGINFO		(xmw->xmaid.msginfo)

#define MSGBUF		(msg->msgbuf)
#define XMW			(msg->xmw)
#define MSGPATH		(msg->msgpath)
#define INPUTFLAG 	(msg->inputflag)
#define INPUTID 	(msg->input_id)
#define FILEDESC	(msg->fd)
#define INPUTBUF	(msg->input_buf)
#define INPUTCUR	(msg->input_cur)
#define INPUTSIZE	(msg->input_size)
#define OUTPUTID 	(msg->output_id)
#define FILEDESCO 	(msg->fdout)
#define OUTPUT  	(msg->output)

XMaidMsgInfo *
xmaidmsginfo_new(XMaidWidget xmw)
{
	XMaidMsgInfo *msg;
	if ((msg = malloc(sizeof(*msg))) == NULL) {
		perror("xmaidmsginfo_new");
		exit(1);
	}

	XMW			= xmw;
	MSGBUF		= NULL;
	MSGPATH		= NULL;
	
	INPUTID     = OUTPUTID = 0;
	FILEDESC    = FILEDESCO = -1;
	INPUTBUF    = NULL;
	INPUTCUR    = INPUTSIZE = 0;
	INPUTFLAG   = 0;
	OUTPUT      = NULL;

	return msg;
}

static void
xmaidmsginfo_clear_output(XMaidMsgInfo *msg)
{
	/* clear old output */
	if (OUTPUTID) {
		XtRemoveInput(OUTPUTID);
		OUTPUTID = 0;
	}
	if (FILEDESCO >= 0) {
		close(FILEDESCO);
		FILEDESCO = -1;
	}
	msgpacket_delete(OUTPUT);
	OUTPUT = NULL;
}

static void
OutputHandler(XtPointer cl, int *fdp, XtInputId *id)
{
	int len;
	XMaidMsgInfo *msg = cl;

	if ((len = write(*fdp, OUTPUT->p, OUTPUT->len)) >= 0) {
		OUTPUT->p += len;
		OUTPUT->len -= len;
		if (OUTPUT->len > 0)
			return;
	} else
		perror("xmaidmsg OutputHandler");

	/* end */
	xmaidmsginfo_clear_output(msg);
}

/* need to select output */
void
xmaid_msginfo_entry_string(XMaidMsgInfo *msg, char *str, int len)
{
	int fds[2];

	if (pipe(fds) < 0) {
		perror("xmaidmsg_string_entry");
		return;
	}

	/* clear old output */
	xmaidmsginfo_clear_output(msg);

	/* entry new output */
	XtVaSetValues(MSGBUF, XtNfileid, fds[0], NULL);
	OUTPUT   = msgpacket_new(str, len);
	OUTPUTID = XtAppAddInput(XtWidgetToApplicationContext(MSGBUF), 
							 FILEDESCO = fds[1],
							 (XtPointer)XtInputWriteMask, 
							 OutputHandler, (XtPointer)msg);
}

static char *
xmaid_make_msgpath(char *oldmsgpath, char *msgpath)
{
	char *buf = NULL, *p, *q;
	int len;

	if (oldmsgpath == NULL)
		len = strlen(oldmsgpath) + 1;
	else
		len = strlen(oldmsgpath) + strlen(msgpath) + 1;
	len += sizeof(DEFAULT_FILE);

	if ((buf = malloc(len))) {
		if (oldmsgpath == NULL || (p = strrchr(oldmsgpath, '/')) == NULL)
			strcpy(buf, msgpath);
		else
			sprintf(buf, "%1.*s%s", p - oldmsgpath+1, oldmsgpath, msgpath);

		p = (*buf == '/')?buf+1:buf;
		for(;;) {
			if (*p == '/') {
				strcpy(buf, p);
				p = buf+1;
			} else if (!strncmp(p, "../", 3)) {
				q = p;
				while (*q != '/') q--;
				strcpy(q, p+2);
				p = q+1;
			} if (!strncmp(p, "./", 2)) {
				strcpy(p, p+2);
			} else if ((q = strchr(p, '/'))) {
				p = q+1;
			} else
				break;
		}
	}
	
#if 0
	fprintf(stderr,"name:%s\n", buf);
#endif
	
	return buf;
}

void
xmaid_msginfo_entry_file(XMaidMsgInfo *msg, char *name)
{
	char *p;

	if (name) {
		struct stat sb;
		if (stat(name, &sb) < 0)
			perror("xmaidmsginfo_entry_file:stat");
		else {
			if ((sb.st_mode & S_IFREG) && 
				(sb.st_mode & S_IXUSR)) /* executable */
				child_exec(name);
			else {
				if (!(p = strrchr(name, '.')) || strncmp(++p, "xhtml", 5)) {
					/* entry new file */
					xmaidmsginfo_clear_output(msg);
					XtVaSetValues(MSGBUF, XtNfile, name,
								  XtNfileid, -1, NULL);
				} else				
					xmaid_msginfo_entry_parsefile(msg, name);

				if (MSGPATH)
					free(MSGPATH);
				MSGPATH = name;
				return;
			}	
		}
		free(name);
		return;
	}
}

/* for function using fork(2)                  */
/* other process doesn't need to select output */
void
xmaid_msginfo_entry_fd(XMaidMsgInfo *msg, int fd, char *name)
{
	xmaidmsginfo_clear_output(msg);
	XtVaSetValues(MSGBUF, XtNfileid, fd, XtNfile, name, NULL);
}

static void
InputHandler(XtPointer cl, int *fdp, XtInputId *id)
{
	XMaidMsgInfo *msg = cl;

	char *q, *r;
	int l;
	
	if (INPUTSIZE - INPUTCUR < 128) {
		char *p;
		INPUTSIZE += 512;
		if ((p = (INPUTBUF)?realloc(INPUTBUF, INPUTSIZE + 1):malloc(INPUTSIZE + 1)) == NULL) {
			perror("xmaidmsg InputHandler");
			exit(1);
		}
		INPUTBUF = p;
	}

	if ((l = read(*fdp, INPUTBUF+INPUTCUR, INPUTSIZE - INPUTCUR)) < 0) {
		perror("retmsg_recv");
		goto end;
	} else if (l == 0)
		goto end;

	INPUTCUR += l;
	*(INPUTBUF + INPUTCUR) = '\0';

	if ((q = strchr(INPUTBUF, '\n'))) {
		*q = '\0';
		r = INPUTBUF;
		if (!strncmp(r,"url:",4)) {
			int mode = atoi(r+4);
			r += 6;
#if 0
			fprintf(stderr,"path[%s] name[%s]\n", MSGPATH, r);
#endif
			if (!strncmp(r, "xmaid:", 6)) {
				r += 6;
				if (!strncmp(r, "end", 3)) {
				  XtAppSetExitFlag(XtWidgetToApplicationContext((Widget)XMW));
					XtDestroyWidget((Widget)XMW);
				} else if (!strncmp(r, "kill", 4)) {
					childinfo_kill(atoi(r + 5));
					xmaid_msginfo_entry_file(msg, strdup(MSGPATH));
				}
			} else {
				if (!mode){
					xmaid_msginfo_entry_file(msg, 
											 xmaid_make_msgpath(MSGPATH,r));
				} else {
					xmaid_msginfo_open3(XMW, 600, 500, 
										xmaid_make_msgpath(MSGPATH,r));
				}
			}
		}
		q++;
		strcpy(INPUTBUF,q);
		INPUTCUR -= (q - INPUTBUF);
	}
	return;

 end:

	/* msgbuf destroyed ... clear input */
	XtRemoveInput(INPUTID);
	INPUTID = 0;
	close(FILEDESC);
	FILEDESC = -1;
	INPUTFLAG = 0;
	if (INPUTBUF)
		free(INPUTBUF);
	INPUTBUF = NULL;
	INPUTSIZE = INPUTCUR = 0;
	MSGBUF    = NULL;

	/* and clear output */
	xmaidmsginfo_clear_output(msg);

	if (msg == XMW->xmaid.msginfo) {
		XMaidStartMove(XMW);
	} else {
		XMaidDeleteMsgInfo(XMW, msg);
		xmaidmsginfo_delete(msg);
	}
}

void
xmaid_msginfo_open(XMaidMsgInfo *msg, Widget msgbuf)
{
	int fds[2];

	msg->msgbuf = msgbuf;
	if (pipe(fds) < 0) {
		perror("xmaidmsginfo_open");
		return;
	}
	XtVaSetValues(MSGBUF, XtNfileidout, fds[1], NULL);
	INPUTID = XtAppAddInput(XtWidgetToApplicationContext(MSGBUF), 
							FILEDESC = fds[0],
							(XtPointer)XtInputReadMask, 
							InputHandler, msg);
	INPUTFLAG = 1;
}

#define MENUX			(xmw->xmaid.menu_x)
#define MENUY			(xmw->xmaid.menu_y)
#define MENUDX			(xmw->xmaid.menu_dx)
#define MENUDY			(xmw->xmaid.menu_dy)


void
xmaid_msginfo_open2(XMaidWidget xmw, u_int w, u_int h, char *name)
{
	Widget msgbuf;
	XMaidStopMove(xmw);
	msgbuf = XtVaCreatePopupShell("popup",
								  msgShellWidgetClass, (Widget)xmw,
								  XtNwidth, w,
								  XtNheight, h,
								  XtNx, xmw->core.x + MENUX,
								  XtNy, xmw->core.y + MENUY,
								  XtNdiffxpos, MENUDX,
								  XtNdiffypos, MENUDY, 
								  XtNoverrideRedirect, True,
								  NULL);

	xmaid_msginfo_open(MSGINFO, msgbuf);
	xmaid_msginfo_entry_file(MSGINFO, strdup(name));
	XtPopup(msgbuf, XtGrabNone);
}

void
xmaid_msginfo_open3(XMaidWidget xmw, u_int w, u_int h, char *name)
{
	Widget msgbuf;
	XMaidMsgInfo *msg;

	msgbuf = XtVaCreateWidget("message",
							  msgShellWidgetClass, (Widget)xmw,
							  XtNwidth, w,
							  XtNheight, h,
							  XtNhukidasi, False,
							  NULL );
	msg = xmaidmsginfo_new(xmw);
	xmaid_msginfo_open(msg, msgbuf);
	XMaidAddMsgInfo(xmw, msg);
	xmaid_msginfo_entry_file(msg, name);
	XtRealizeWidget(msgbuf);
	XtMapWidget(msgbuf);
}

/* XXX almost same in MsgShell.c  It shuld be marge */
static void
CloseHandler(XtPointer cl, int *fdp, XtInputId *id)
{
	int len;
	char buf[100];
	if ((len = read(*fdp, buf, sizeof(buf))) < 0) {
		perror("CLoseHandler");
		XtRemoveInput(*id);
		close(*fdp);
	} else if (len == 0) {
		XtRemoveInput(*id);
		close(*fdp);
	}
}

static void 
MyClose(Widget widget, int fd)
{
	XtAppAddInput(XtWidgetToApplicationContext(widget),
				  fd, (XtPointer)XtInputReadMask,
				  CloseHandler, NULL);
}

void
xmaidmsginfo_delete(XMaidMsgInfo *msg)
{
	if (msg) {
		xmaidmsginfo_clear_output(msg);
		if (INPUTID)
			XtRemoveInput(INPUTID);
		if (FILEDESC >= 0)
			MyClose((Widget)XMW, FILEDESC);
		if (INPUTBUF)
			free(INPUTBUF);
		if (MSGBUF)
			XtDestroyWidget(MSGBUF);
		free(msg);
	}
}


void
xmaidmsginfolist_delete(XMaidMsgInfoList l)
{
	XMaidMsgInfo *p = l, *q;
	while (p != NULL) {
		q = p;
		p = p->next;
		xmaidmsginfo_delete(q);
	}
}
